from sqlalchemy import and_, func, join, select
from sqlalchemy_utils.view import create_view

import vigorish.database as db
from vigorish.data.metrics.bat_stats import BatStatsMetrics
from vigorish.views.bat_stats_col_expressions import (
    avg,
    bb_rate,
    contact_rate,
    iso,
    k_rate,
    obp,
    ops,
    slg,
)


class Team_BatStats_By_Year_View(db.Base):
    __table__ = create_view(
        name="team_batstats_by_year",
        selectable=select(
            [
                db.BatStats.player_team_id.label("team_id"),
                db.BatStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                func.count(db.BatStats.game_status_id.distinct()).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
                db.BatStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.BatStats, db.Season, db.BatStats.season_id == db.Season.id))
        .group_by(db.BatStats.season_id)
        .group_by(db.BatStats.player_team_id)
        .order_by(db.BatStats.player_team_id)
        .order_by(db.BatStats.season_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results)[0] if results else None

    @classmethod
    def get_bat_stats_by_year_for_team(cls, db_engine, team_id_bbref):
        s = select([cls]).where(cls.team_id_bbref == team_id_bbref)
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []

    @classmethod
    def get_bat_stats_for_season_for_all_teams(cls, db_engine, year):
        s = select([cls]).where(cls.year == year)
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []


class Team_BatStats_By_BatOrder_By_Year(db.Base):
    __table__ = create_view(
        name="team_batstats_by_batorder_by_year",
        selectable=select(
            [
                db.BatStats.player_team_id.label("team_id"),
                db.BatStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                db.BatStats.bat_order.label("bat_order"),
                func.count(db.BatStats.game_status_id.distinct()).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
                db.BatStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.BatStats, db.Season, db.BatStats.season_id == db.Season.id))
        .group_by(db.BatStats.season_id)
        .group_by(db.BatStats.player_team_id)
        .group_by(db.BatStats.bat_order)
        .order_by(db.BatStats.bat_order),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_by_lineup_spot_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []

    @classmethod
    def get_bat_stats_for_lineup_spot_by_year_for_team(cls, db_engine, lineup_spot, team_id_bbref):
        s = select([cls]).where(and_(cls.bat_order == lineup_spot, cls.team_id_bbref == team_id_bbref))
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []

    @classmethod
    def get_bat_stats_for_lineup_spot_for_season_for_all_teams(cls, db_engine, lineup_spot, year):
        s = select([cls]).where(and_(cls.bat_order == lineup_spot, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []


class Team_BatStats_By_DefPosition_By_Year(db.Base):
    __table__ = create_view(
        name="team_batstats_by_defpos_by_year",
        selectable=select(
            [
                db.BatStats.player_team_id.label("team_id"),
                db.BatStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                db.BatStats.def_position.label("def_position"),
                func.count(db.BatStats.game_status_id.distinct()).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
                db.BatStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.BatStats, db.Season, db.BatStats.season_id == db.Season.id))
        .group_by(db.BatStats.season_id)
        .group_by(db.BatStats.player_team_id)
        .group_by(db.BatStats.def_position)
        .order_by(db.BatStats.def_position),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_by_defpos_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []

    @classmethod
    def get_bat_stats_for_defpos_by_year_for_team(cls, db_engine, def_position, team_id_bbref):
        s = select([cls]).where(and_(cls.def_position == def_position, cls.team_id_bbref == team_id_bbref))
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []

    @classmethod
    def get_bat_stats_for_defpos_for_season_for_all_teams(cls, db_engine, def_position, year):
        s = select([cls]).where(and_(cls.def_position == def_position, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []


class Team_BatStats_For_Starters_By_Year(db.Base):
    __table__ = create_view(
        name="team_batstats_for_starters_by_year",
        selectable=select(
            [
                db.BatStats.player_team_id.label("team_id"),
                db.BatStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                db.BatStats.is_starter.label("is_starter"),
                func.count(db.BatStats.game_status_id.distinct()).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
                db.BatStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.BatStats, db.Season, db.BatStats.season_id == db.Season.id))
        .where(db.BatStats.is_starter == 1)
        .group_by(db.BatStats.season_id)
        .group_by(db.BatStats.player_team_id)
        .order_by(db.BatStats.player_team_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_for_starters_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results)[0] if results else None

    @classmethod
    def get_bat_stats_for_starters_by_year_for_team(cls, db_engine, team_id_bbref):
        s = select([cls]).where(cls.team_id_bbref == team_id_bbref)
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []

    @classmethod
    def get_bat_stats_for_starters_for_season_for_all_teams(cls, db_engine, year):
        s = select([cls]).where(cls.year == year)
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []


class Team_BatStats_For_Subs_By_Year(db.Base):
    __table__ = create_view(
        name="team_batstats_for_subs_by_year",
        selectable=select(
            [
                db.BatStats.player_team_id.label("team_id"),
                db.BatStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                db.BatStats.is_starter.label("is_starter"),
                func.count(db.BatStats.game_status_id.distinct()).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
                db.BatStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.BatStats, db.Season, db.BatStats.season_id == db.Season.id))
        .where(db.BatStats.is_starter == 0)
        .group_by(db.BatStats.season_id)
        .group_by(db.BatStats.player_team_id)
        .order_by(db.BatStats.player_team_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_for_subs_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results)[0] if results else None

    @classmethod
    def get_bat_stats_for_subs_by_year_for_team(cls, db_engine, team_id_bbref):
        s = select([cls]).where(cls.team_id_bbref == team_id_bbref)
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []

    @classmethod
    def get_bat_stats_for_subs_for_season_for_all_teams(cls, db_engine, year):
        s = select([cls]).where(cls.year == year)
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []


class Team_BatStats_By_Player_By_Year_View(db.Base):
    __table__ = create_view(
        name="team_batstats_by_player_by_year",
        selectable=select(
            [
                db.BatStats.player_team_id.label("team_id"),
                db.BatStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                db.BatStats.player_id_mlb.label("mlb_id"),
                db.BatStats.player_id_bbref.label("bbref_id"),
                func.count(db.BatStats.game_status_id.distinct()).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
                db.BatStats.player_id.label("player_id"),
                db.BatStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.BatStats, db.Season, db.BatStats.season_id == db.Season.id))
        .group_by(db.BatStats.season_id)
        .group_by(db.BatStats.player_id)
        .group_by(db.BatStats.player_team_id)
        .order_by(db.BatStats.plate_appearances),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_by_player_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []


class Team_BatStats_By_BatOrder_By_Player_By_Year(db.Base):
    __table__ = create_view(
        name="team_batstats_by_batorder_by_player_by_year",
        selectable=select(
            [
                db.BatStats.player_team_id.label("team_id"),
                db.BatStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                db.BatStats.bat_order.label("bat_order"),
                db.BatStats.player_id_mlb.label("mlb_id"),
                db.BatStats.player_id_bbref.label("bbref_id"),
                func.count(db.BatStats.game_status_id.distinct()).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
                db.BatStats.player_id.label("player_id"),
                db.BatStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.BatStats, db.Season, db.BatStats.season_id == db.Season.id))
        .group_by(db.BatStats.player_team_id)
        .group_by(db.BatStats.player_id)
        .group_by(db.BatStats.bat_order)
        .order_by(db.BatStats.bat_order),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_for_lineup_spot_by_player_for_team(cls, db_engine, lineup_spot, team_id_bbref, year):
        s = select([cls]).where(
            and_(cls.bat_order == lineup_spot, cls.team_id_bbref == team_id_bbref, cls.year == year)
        )
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []


class Team_BatStats_By_DefPosition_By_Player_By_Year(db.Base):
    __table__ = create_view(
        name="team_batstats_by_defpos_by_player_by_year",
        selectable=select(
            [
                db.BatStats.player_team_id.label("team_id"),
                db.BatStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                db.BatStats.def_position.label("def_position"),
                db.BatStats.player_id_mlb.label("mlb_id"),
                db.BatStats.player_id_bbref.label("bbref_id"),
                func.count(db.BatStats.game_status_id.distinct()).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
                db.BatStats.player_id.label("player_id"),
                db.BatStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.BatStats, db.Season, db.BatStats.season_id == db.Season.id))
        .group_by(db.BatStats.player_team_id)
        .group_by(db.BatStats.player_id)
        .group_by(db.BatStats.def_position)
        .order_by(db.BatStats.def_position),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_for_defpos_by_player_for_team(cls, db_engine, def_position, team_id_bbref, year):
        s = select([cls]).where(
            and_(cls.def_position == def_position, cls.team_id_bbref == team_id_bbref, cls.year == year)
        )
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []


class Team_BatStats_For_Starters_By_Player_By_Year(db.Base):
    __table__ = create_view(
        name="team_batstats_for_starters_by_player_by_year",
        selectable=select(
            [
                db.BatStats.player_team_id.label("team_id"),
                db.BatStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                db.BatStats.is_starter.label("is_starter"),
                db.BatStats.player_id_mlb.label("mlb_id"),
                db.BatStats.player_id_bbref.label("bbref_id"),
                func.count(db.BatStats.game_status_id.distinct()).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
                db.BatStats.player_id.label("player_id"),
                db.BatStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.BatStats, db.Season, db.BatStats.season_id == db.Season.id))
        .where(db.BatStats.is_starter == 1)
        .group_by(db.BatStats.player_team_id)
        .group_by(db.BatStats.player_id)
        .group_by(db.BatStats.def_position)
        .order_by(db.BatStats.def_position),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_for_starters_by_player_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []


class Team_BatStats_For_Subs_By_Player_By_Year(db.Base):
    __table__ = create_view(
        name="team_batstats_for_subs_by_player_by_year",
        selectable=select(
            [
                db.BatStats.player_team_id.label("team_id"),
                db.BatStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                db.BatStats.is_starter.label("is_starter"),
                db.BatStats.player_id_mlb.label("mlb_id"),
                db.BatStats.player_id_bbref.label("bbref_id"),
                func.count(db.BatStats.game_status_id.distinct()).label("total_games"),
                avg,
                obp,
                slg,
                ops,
                iso,
                bb_rate,
                k_rate,
                contact_rate,
                func.sum(db.BatStats.plate_appearances).label("plate_appearances"),
                func.sum(db.BatStats.at_bats).label("at_bats"),
                func.sum(db.BatStats.hits).label("hits"),
                func.sum(db.BatStats.runs_scored).label("runs_scored"),
                func.sum(db.BatStats.rbis).label("rbis"),
                func.sum(db.BatStats.bases_on_balls).label("bases_on_balls"),
                func.sum(db.BatStats.strikeouts).label("strikeouts"),
                func.sum(db.BatStats.doubles).label("doubles"),
                func.sum(db.BatStats.triples).label("triples"),
                func.sum(db.BatStats.homeruns).label("homeruns"),
                func.sum(db.BatStats.stolen_bases).label("stolen_bases"),
                func.sum(db.BatStats.caught_stealing).label("caught_stealing"),
                func.sum(db.BatStats.hit_by_pitch).label("hit_by_pitch"),
                func.sum(db.BatStats.intentional_bb).label("intentional_bb"),
                func.sum(db.BatStats.gdp).label("gdp"),
                func.sum(db.BatStats.sac_fly).label("sac_fly"),
                func.sum(db.BatStats.sac_hit).label("sac_hit"),
                func.sum(db.BatStats.total_pitches).label("total_pitches"),
                func.sum(db.BatStats.total_strikes).label("total_strikes"),
                func.sum(db.BatStats.wpa_bat).label("wpa_bat"),
                func.sum(db.BatStats.wpa_bat_pos).label("wpa_bat_pos"),
                func.sum(db.BatStats.wpa_bat_neg).label("wpa_bat_neg"),
                func.sum(db.BatStats.re24_bat).label("re24_bat"),
                db.BatStats.player_id.label("player_id"),
                db.BatStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.BatStats, db.Season, db.BatStats.season_id == db.Season.id))
        .where(db.BatStats.is_starter == 0)
        .group_by(db.BatStats.player_team_id)
        .group_by(db.BatStats.player_id)
        .group_by(db.BatStats.def_position)
        .order_by(db.BatStats.def_position),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_bat_stats_for_subs_by_player_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return BatStatsMetrics.from_query_results(results) if results else []



from urllib.parse import urljoin

from lxml import html

from vigorish.scrape.bbref_games_for_date.models.game_info import BBRefGameInfo
from vigorish.scrape.bbref_games_for_date.models.games_for_date import BBRefGamesForDate
from vigorish.util.dt_format_strings import DATE_ONLY, DATE_ONLY_2
from vigorish.util.result import Result
from vigorish.util.string_helpers import get_bbref_game_id_from_url, validate_bbref_game_id

XPATH_BOXSCORE_URL_MAIN_CONTENT = (
    '//div[@id="content"]//div[contains(@class, "game_summaries")]' '//td[contains(@class, "gamelink")]/a/@href'
)

XPATH_BOXSCORE_URL_HEADER_NAV = (
    '//li[@id="header_scores"]//div[contains(@class, "game_summaries")]' '//td[contains(@class, "gamelink")]/a/@href'
)


def parse_bbref_dashboard_page(scraped_html, game_date, url):
    page_content = html.fromstring(scraped_html, base_url=url)
    games_for_date = BBRefGamesForDate()
    games_for_date.game_date = game_date
    games_for_date.game_date_str = game_date.strftime(DATE_ONLY)
    games_for_date.dashboard_url = url
    boxscore_urls = page_content.xpath(XPATH_BOXSCORE_URL_MAIN_CONTENT)
    if not boxscore_urls:
        games_for_date.game_count = 0
        return Result.Ok(games_for_date)
    result = verify_boxscore_date(boxscore_urls, game_date, url)
    if result.failure:
        return result
    games_for_date.games = []
    for rel_url in boxscore_urls:
        if "allstar" in rel_url:
            continue
        boxscore_url = urljoin(url, rel_url)
        game_id = get_bbref_game_id_from_url(boxscore_url)
        game_info = BBRefGameInfo(url=boxscore_url, bbref_game_id=game_id)
        games_for_date.games.append(game_info)
    games_for_date.game_count = len(games_for_date.games)
    return Result.Ok(games_for_date)


def verify_boxscore_date(boxscore_urls, game_date, url):
    rel_url = boxscore_urls[0]
    if "allstar" in rel_url:
        return Result.Ok()
    box_url = urljoin(url, rel_url)
    game_id = get_bbref_game_id_from_url(box_url)
    if not game_id:
        return Result.Fail(f"Failed to parse game id from url: {box_url}")
    result = validate_bbref_game_id(game_id)
    if result.failure:
        return result
    game_date_id = result.value["game_date"]
    if game_date_id != game_date.date():
        scrape_date_str = game_date.strftime(DATE_ONLY_2)
        error = (
            f"BBref daily dashboard URL for {scrape_date_str} redirected to game results for the "
            f"previous day. Please try again when boxscores for {scrape_date_str} are available."
        )
        return Result.Fail(error)
    return Result.Ok()



"""Menu that allows the user to view and modify environment variables."""
import subprocess
from sys import exit

from bullet import colors, Input
from getch import pause

from vigorish.cli.components import print_message, yes_no_cancel_prompt, yes_no_prompt
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS
from vigorish.util.result import Result

RESTART_WARNING = "\nApplication must be restarted for these changes to take effect!"


class ChangeEnvVarSetting(MenuItem):
    def __init__(self, app, setting_name):
        super().__init__(app)
        self.menu_item_text = setting_name
        self.menu_item_emoji = EMOJIS.get("SPIRAL")
        self.setting_name = setting_name
        self.current_setting = self.dotenv.get_current_value(setting_name)
        self.restart_required = self.dotenv.restart_required_on_change(setting_name)
        self.exit_menu = False

    def launch(self):
        subprocess.run(["clear"])
        print_message(f"Variable Name: {self.setting_name}\n", fg="bright_magenta", bold=True)
        print_message(f"Current Value: {self.current_setting}\n", fg="bright_yellow", bold=True)
        if not yes_no_prompt(prompt="\nChange current setting?"):
            return Result.Ok(self.exit_menu)
        user_confirmed, new_value = False, None
        while not user_confirmed:
            subprocess.run(["clear"])
            prompt = f"Enter a new value for {self.setting_name}:\n"
            new_value = Input(prompt, word_color=colors.foreground["default"]).launch()
            result = self.confirm_new_value(new_value)
            if result.failure:
                return Result.Ok(self.exit_menu)
            user_confirmed = result.value
        result = self.dotenv.change_value(self.setting_name, new_value)
        if not self.restart_required:
            return result
        print_message(RESTART_WARNING, fg="bright_magenta", bold=True)
        pause(message="Press any key to continue...")
        exit(0)

    def confirm_new_value(self, new_value):
        prompt = (
            f"\nUpdate {self.setting_name} to the value below?"
            f"\nCurrent Value..: {self.current_setting}"
            f"\nNew Value......: {new_value}"
        )
        return yes_no_cancel_prompt(prompt, wrap=False)



"""Boxscore URLs for all games that took place on a single day."""
import json
from dataclasses import dataclass, field
from datetime import datetime
from typing import List

from vigorish.scrape.brooks_games_for_date.models.game_info import BrooksGameInfo
from vigorish.util.list_helpers import as_dict_list, flatten_list2d


@dataclass
class BrooksGamesForDate:
    """Boxscore URLs for all games that took place on a single day."""

    dashboard_url: str = ""
    game_date: datetime = None
    game_date_str: str = ""
    game_count: str = "0"
    games: List[BrooksGameInfo] = field(default_factory=list)

    @property
    def all_pitch_app_ids_for_date(self):
        return sorted(flatten_list2d([g.all_pitch_app_ids for g in self.games]))

    @property
    def all_bbref_game_ids(self):
        return [game.bbref_game_id for game in self.games]

    def as_dict(self):
        """Convert daily boxscore URL values to a dictionary."""
        return {
            "__brooks_games_for_date__": True,
            "dashboard_url": self.dashboard_url,
            "game_date_str": self.game_date_str,
            "game_count": int(self.game_count),
            "games": as_dict_list(self.games),
        }

    def as_json(self):
        """Convert daily game list to JSON."""
        return json.dumps(self.as_dict(), indent=2)



PFX_PLATE_DISCIPLINE_METRICS = {
    "Z%": ("zone_rate", "total_inside_strike_zone"),
    "SS%": ("swinging_strike_rate", "total_swinging_strikes"),
    "O-SW%": ("o_swing_rate", "total_swings_outside_zone"),
    "CON%": ("contact_rate", "total_swings_made_contact"),
}

PFX_BATTED_BALL_METRICS = {
    "GB%": ("ground_ball_rate", "total_ground_balls"),
    "LD%": ("line_drive_rate", "total_line_drives"),
    "FB%": ("fly_ball_rate", "total_fly_balls"),
    "PU%": ("pop_up_rate", "total_pop_ups"),
}



"""Menu item that allows the user to initialize/reset the database."""
import subprocess
from tempfile import TemporaryDirectory

from getch import pause
from halo import Halo

from vigorish.cli.components import print_heading, print_message, yes_no_prompt
from vigorish.cli.components.util import get_random_cli_color, get_random_dots_spinner
from vigorish.cli.menu_item import MenuItem
from vigorish.config.project_paths import NIGHTMAREJS_FOLDER, NODEJS_INBOX
from vigorish.constants import EMOJIS
from vigorish.util.result import Result
from vigorish.util.sys_helpers import node_is_installed, node_modules_folder_exists, run_command

INSTALL_ERROR = "Error! Node.js is not installed, see README for install instructions."
INSTALL_MESSAGE = "You must install all required node packages in order to scrape any data."
INSTALL_PROMPT = (
    "\nSelect YES install Nightmare, Xfvb, and other node packages\n" "Select NO to return to the previous menu"
)
UPDATE_MESSAGE = "Nightmare and all node dependencies are installed."
UPDATE_PROMPT = (
    "\nSelect YES to update all installed packages to the newest version.\n" "Select NO to return to the previous menu."
)


class NpmInstallUpdate(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.menu_heading = "Update Node Packages" if node_modules_folder_exists() else "Install Node Packages"
        self.menu_item_text = self.menu_heading
        self.menu_item_emoji = EMOJIS.get("PACKAGE")

    def launch(self):
        subprocess.run(["clear"])
        print_heading(self.menu_heading, fg="bright_yellow")
        if not node_is_installed():
            print_message(INSTALL_ERROR, fg="bright_red", bold=True)
            pause(message="Press any key to continue...")
            return
        if not NODEJS_INBOX.exists():
            NODEJS_INBOX.mkdir(parents=True, exist_ok=True)
        if node_modules_folder_exists():
            message = UPDATE_MESSAGE
            prompt = UPDATE_PROMPT
            temp_folder = None
            command = "npm update --timeout=9999999"
        else:
            message = INSTALL_MESSAGE
            prompt = INSTALL_PROMPT
            temp_folder = TemporaryDirectory(dir=NIGHTMAREJS_FOLDER)
            command = f"npm install --timeout=9999999 --cache={temp_folder.name}"
        print_message(message, fg="bright_yellow")
        if not yes_no_prompt(prompt, wrap=False):
            return Result.Ok(self.exit_menu)
        subprocess.run(["clear"])
        print_heading(self.menu_heading, fg="bright_yellow")
        spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        spinner.text = "Updating node packages..." if node_modules_folder_exists() else "Installing node packages..."
        spinner.start()
        cmd_output = list(run_command(command, cwd=str(NIGHTMAREJS_FOLDER)))

        subprocess.run(["clear"])
        print_heading(self.menu_heading, fg="bright_yellow")
        spinner.succeed(
            "All node packages are up-to-date!"
            if node_modules_folder_exists()
            else "Successfully installed all node dependencies!"
        )
        for s in cmd_output:
            print_message(s)
        if temp_folder:
            temp_folder.cleanup()
        pause(message="\nPress any key to continue...")
        return Result.Ok(self.exit_menu)



import vigorish.database as db
from vigorish.enums import DataSet, ScrapeCondition
from vigorish.scrape.brooks_games_for_date.parse_html import parse_brooks_dashboard_page
from vigorish.scrape.scrape_task import ScrapeTaskABC
from vigorish.status.update_status_brooks_games_for_date import (
    update_brooks_games_for_date_single_date,
)
from vigorish.util.dt_format_strings import DATE_ONLY
from vigorish.util.result import Result


class ScrapeBrooksGamesForDate(ScrapeTaskABC):
    def __init__(self, app, db_job):
        self.data_set = DataSet.BROOKS_GAMES_FOR_DATE
        super().__init__(app, db_job)

    def check_prerequisites(self, game_date):
        bbref_games_for_date = db.DateScrapeStatus.verify_bbref_daily_dashboard_scraped_for_date(
            self.db_session, game_date
        )
        if bbref_games_for_date:
            return Result.Ok()
        date_str = game_date.strftime(DATE_ONLY)
        error = (
            f"BBref games for date ({date_str}) have not been scraped, unable to scrape Brooks "
            f"games for date ({date_str}) until this has been done."
        )
        return Result.Fail(error)

    def check_current_status(self, game_date):
        if self.scrape_condition == ScrapeCondition.ALWAYS:
            return Result.Ok()
        brooks_games_for_date = db.DateScrapeStatus.verify_brooks_daily_dashboard_scraped_for_date(
            self.db_session, game_date
        )
        return Result.Ok() if not brooks_games_for_date else Result.Fail("skip")

    def parse_html(self, url_details):
        bbref_games_for_date = self.scraped_data.get_bbref_games_for_date(url_details.url_id)
        if not bbref_games_for_date:
            return Result.Fail(f"Failed to retrieve {self.data_set} (URL ID: {url_details.url_id})")
        return parse_brooks_dashboard_page(
            self.db_session,
            url_details.html,
            url_details.url_id,
            url_details.url,
            bbref_games_for_date,
        )

    def update_status(self, parsed_data):
        return update_brooks_games_for_date_single_date(self.db_session, self.season, parsed_data)



"""Menu that allows the user to view all jobs grouped by status."""
import subprocess

from vigorish.cli.components import user_options_prompt
from vigorish.cli.menu import Menu
from vigorish.cli.menu_items.investigate_data_failures import InvestigateScrapedDataFailures
from vigorish.cli.menu_items.investigate_invalid_pfx import InvestigateInvalidPitchFx
from vigorish.cli.menu_items.investigate_pitchfx_errors import InvestigatePitchFxErrors
from vigorish.cli.menu_items.return_to_parent import ReturnToParent
from vigorish.constants import EMOJIS, MENU_NUMBERS
from vigorish.util.result import Result


class ScrapedDataErrorsMenu(Menu):
    def __init__(self, app):
        super().__init__(app)
        self.menu_text = "Select the type of error to investigate:"
        self.menu_item_text = "Investigate Failures"
        self.menu_item_emoji = EMOJIS.get("FLASHLIGHT")

    def launch(self):
        exit_menu = False
        while not exit_menu:
            subprocess.run(["clear"])
            result = self.audit_report_season_prompt()
            subprocess.run(["clear"])
            if result.failure:
                return Result.Ok(exit_menu)
            self.year = result.value
            self.season_report = self.app.audit_report[self.year]
            self.populate_menu_items()
            result = self.prompt_user_for_menu_selection()
            if result.failure:
                return result
            exit_menu = result.value
            self.menu_items = []
        return Result.Ok(exit_menu)

    def populate_menu_items(self):
        self.menu_items.clear()
        failed_gids = self.season_report.get("failed", [])
        if failed_gids:
            self.menu_items.append(InvestigateScrapedDataFailures(self.app, self.year, failed_gids))
        pfx_error_gids = self.season_report.get("pfx_error", [])
        if pfx_error_gids:
            self.menu_items.append(InvestigatePitchFxErrors(self.app, self.year, pfx_error_gids))
        invalid_pfx_gids = self.season_report.get("invalid_pfx", [])
        if invalid_pfx_gids:
            self.menu_items.append(InvestigateInvalidPitchFx(self.app, self.year, invalid_pfx_gids))
        self.menu_items.append(ReturnToParent(self.app, "Return to Main Menu"))

    def audit_report_season_prompt(self):
        prompt = "Select an MLB season from the list below:"
        years_with_errors = [
            year
            for year in self.app.audit_report.keys()
            if self.app.audit_report[year].get("invalid_pfx", [])
            or self.app.audit_report[year].get("pfx_error", [])
            or self.app.audit_report[year].get("failed", [])
        ]
        choices = {f"{MENU_NUMBERS.get(num)}  {year}": year for num, year in enumerate(years_with_errors, start=1)}
        choices[f"{EMOJIS.get('BACK')} Return to Previous Menu"] = None
        return user_options_prompt(choices, prompt)



"""Biographical information for a single player."""
from sqlalchemy import Boolean, Column, DateTime, Integer, String
from sqlalchemy.orm import backref, relationship

import vigorish.database as db


class Player(db.Base):
    """Biographical information for a single player."""

    __tablename__ = "player"
    id = Column(Integer, primary_key=True)
    name_first = Column(String)
    name_last = Column(String)
    name_given = Column(String)
    bats = Column(String)
    throws = Column(String)
    weight = Column(Integer)
    height = Column(Integer)
    debut = Column(DateTime)
    birth_year = Column(Integer)
    birth_month = Column(Integer)
    birth_day = Column(Integer)
    birth_country = Column(String)
    birth_state = Column(String)
    birth_city = Column(String)
    bbref_id = Column(String, index=True)
    retro_id = Column(String, default="")
    mlb_id = Column(Integer, index=True)
    scraped_transactions = Column(Boolean, default=False)
    minor_league_player = Column(Boolean, default=False)
    missing_mlb_id = Column(Boolean, default=True)

    id_map = relationship("PlayerId", backref=backref("player", uselist=False))
    pitch_apps = relationship("PitchAppScrapeStatus", backref="player")
    teams = relationship("Team", secondary="player_team", back_populates="players")

    def __repr__(self):
        return f"<Player name={self.name_first} {self.name_last}, bbref_id={self.bbref_id}>"

    @classmethod
    def find_by_bbref_id(cls, db_session, bbref_id):
        return db_session.query(cls).filter_by(bbref_id=bbref_id).first()

    @classmethod
    def find_by_mlb_id(cls, db_session, mlb_id) -> "Player":
        return db_session.query(cls).filter_by(mlb_id=mlb_id).first()



"""CLI elements that extend Bullet classes for"""
import subprocess

from bullet import colors, Input
from getch import pause

from vigorish.cli.components.util import print_error, print_heading, print_message
from vigorish.util.regex import JOB_NAME_REGEX

DEFAULT_JOB_NAME = "use default"


class JobNameInput(Input):
    """
    Custom Input element that accepts strings consisting of ONLY a-z. A-Z, 0-9, hyphen ("-")
    and/or underscore ("_") characters. Also, user is allowed to provide no value and dismiss
    the prompt, since the JobName value the name refers to is an optional value on the
    SQLAlchemy model for the ScrapeJob class.
    """

    def __init__(
        self,
        prompt,
        heading=None,
        heading_color=None,
        message=None,
        default=None,
        indent=0,
        word_color=colors.foreground["default"],
        strip=False,
        pattern="",
    ):
        self.heading = heading
        self.heading_color = heading_color
        self.message = message
        if not default:
            default = DEFAULT_JOB_NAME
        super().__init__(prompt, default, indent, word_color, strip, pattern)

    def launch(self):
        while True:
            subprocess.run(["clear"])
            if self.heading:
                print_heading(self.heading, fg=self.heading_color)
            if self.message:
                print_message(f"{self.message}\n")
            result = super().launch()
            if result == DEFAULT_JOB_NAME:
                return None
            if JOB_NAME_REGEX.match(result):
                return result
            error = (
                f"\n'{result}' contains one or more invalid characters. Job name must contain "
                "only letters, numbers, hyphen and underscore characters.\n"
            )
            print_error(error, fg="bright_red")
            pause(message="Press any key to continue...")



"""CLI application entry point."""
import os
import subprocess

import click
from getch import pause

from vigorish.app import Vigorish
from vigorish.cli.click_params import DataSetName, DateString, FileTypeName, JobName, MlbSeason
from vigorish.cli.components import print_message, validate_scrape_dates
from vigorish.cli.main_menu import MainMenu
from vigorish.config.project_paths import VIG_FOLDER
from vigorish.enums import DataSet, StatusReport, SyncDirection
from vigorish.scrape.job_runner import JobRunner
from vigorish.status.report_status import (
    report_date_range_status,
    report_season_status,
    report_status_single_date,
)
from vigorish.tasks import SyncDataNoPromptsTask
from vigorish.util.datetime_util import current_year, today_str
from vigorish.util.result import Result
from vigorish.util.string_helpers import flatten_list2d


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.pass_context
def cli(ctx):
    """
    Vigorish scrapes various websites for MLB data.

    Please visit https://aaronluna.dev/projects/vigorish for user guides and project documentation.
    """
    if os.environ.get("ENV") != "TEST":  # pragma: no cover
        if not VIG_FOLDER.exists():
            VIG_FOLDER.mkdir()
    ctx.obj = Vigorish()


@cli.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.pass_obj
def ui(app):  # pragma: no cover
    """Menu-driven UI powered by Bullet."""
    try:
        result = MainMenu(app).launch()
        return exit_app(app, result)
    except Exception as e:
        return exit_app(app, Result.Fail(f"Error: {repr(e)}"))


@cli.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.confirmation_option(prompt="Are you sure you want to delete all existing data?")
@click.pass_obj
def setup(app):  # pragma: no cover
    """Populate database with initial player, team and season data.

    WARNING! Before the setup process begins, all existing data will be deleted. This cannot be undone.
    """
    print()  # place an empty line between the command and the progress bars
    result = app.initialize_database()
    return exit_app(app, result, "Successfully populated database with initial data.")


@cli.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "--data-set",
    type=DataSetName(),
    multiple=True,
    default=[str(DataSet.ALL)],
    show_default=True,
    help="Data set(s) to scrape, multiple values can be provided.",
)
@click.option(
    "--start",
    type=DateString(),
    prompt=True,
    help=("Date to start scraping data, string can be in any format that is recognized by dateutil.parser."),
)
@click.option(
    "--end",
    type=DateString(),
    prompt=True,
    help=("Date to stop scraping data, string can be in any format that is recognized by dateutil.parser."),
)
@click.option(
    "--job-name",
    type=JobName(),
    default="",
    help="A name to help identify this job.",
    prompt=("(Optional) Enter a name for this job (ONLY letters, numbers, underscore, and/or hyphen characters)"),
)
@click.pass_obj
def scrape(app, data_set, start, end, job_name):
    """Scrape MLB data from websites."""
    result = validate_scrape_dates(app.db_session, start, end)
    if result.failure:
        return exit_app(app, result)
    new_scrape_job = app.create_scrape_job(data_set, start, end, job_name).value
    job_runner = JobRunner(app=app, db_job=new_scrape_job)
    result = job_runner.execute()
    return exit_app(app, result)


@cli.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.pass_obj
def status(app):
    """Report progress of scraped data, by date or MLB season."""
    pass


@status.command("date", context_settings={"help_option_names": ["-h", "--help"]})
@click.argument("game_date", type=DateString(), default=today_str)
@click.option(
    "-v",
    "verbosity",
    count=True,
    default=1,
    help=(
        "Specify the level of detail to report:\n"
        "    -v: report the combined scrape progress for all games on the specified date.\n"
        "   -vv: report combined and individual scrape progress for each game on the specified date\n"
        "  -vvv: report combined/individual game scrape progress and pitch appearance scrape progress\n"
    ),
)
@click.pass_obj
def status_date(app, game_date, verbosity):
    """Report status for a single date.

    Dates can be provided in any format that is recognized by dateutil.parser.
    For example, all of the following strings are valid ways to represent the same date:
    "2018-5-13" -or- "05/13/2018" -or- "May 13 2018"
    """
    report_type = StatusReport.NONE
    if verbosity <= 0:
        error = f"Invalid value for verbosity: {verbosity}. Value must be greater than zero."
        return exit_app(app, Result.Fail(error))
    elif verbosity == 1:
        report_type = StatusReport.DATE_DETAIL_ALL_DATES
    elif verbosity == 2:
        report_type = StatusReport.DATE_DETAIL_MISSING_PITCHFX
    else:
        report_type = StatusReport.SINGLE_DATE_WITH_GAME_STATUS
    result = report_status_single_date(app.db_session, game_date, report_type)
    if result.success:
        report_viewer = result.value
        report_viewer.launch()
    return exit_app(app, result)


@status.command("range", context_settings={"help_option_names": ["-h", "--help"]})
@click.option("--start", type=DateString(), prompt=True, help="First date to report status.")
@click.option("--end", type=DateString(), prompt=True, help="Last date to report status.")
@click.option(
    "-v",
    "verbosity",
    count=True,
    default=1,
    help=(
        "Specify the level of detail to report:\n"
        "    -v: summary report of only dates missing data\n"
        "   -vv: summary report of all dates\n"
        "  -vvv: detailed report of only dates missing data\n"
        " -vvvv: detailed report of all dates\n"
        "-vvvvv: detailed report of all dates with missing pitch_app_ids"
    ),
)
@click.pass_obj
def status_date_range(app, start, end, verbosity):
    """Report status for each date in a specified range.

    Dates can be provided in any format that is recognized by dateutil.parser.
    For example, all of the following strings are valid ways to represent the same date:
    "2018-5-13" -or- "05/13/2018" -or- "May 13 2018"
    """
    report_type = StatusReport.NONE
    if verbosity <= 0:
        error = f"Invalid value for verbosity: {verbosity}. Value must be greater than zero."
        return exit_app(app, Result.Fail(error))
    elif verbosity == 1:
        report_type = StatusReport.DATE_SUMMARY_MISSING_DATA
    elif verbosity == 2:
        report_type = StatusReport.DATE_SUMMARY_ALL_DATES
    elif verbosity == 3:
        report_type = StatusReport.DATE_DETAIL_MISSING_DATA
    elif verbosity == 4:
        report_type = StatusReport.DATE_DETAIL_ALL_DATES
    else:
        report_type = StatusReport.DATE_DETAIL_MISSING_PITCHFX
    result = report_date_range_status(app.db_session, start, end, report_type)
    if result.success:
        report_viewer = result.value
        report_viewer.launch()
    return exit_app(app, result)


@status.command("season", context_settings={"help_option_names": ["-h", "--help"]})
@click.argument("year", type=MlbSeason(), default=current_year)
@click.option(
    "-v",
    "verbosity",
    count=True,
    default=1,
    help=(
        "Specify the level of detail to report:"
        "     -v: overall metrics for entire season"
        "    -vv: summary report of dates in season that are missing data"
        "   -vvv: summary report of all dates in season"
        "  -vvvv: detailed report of dates in season that are missing data"
        "  -vvvv: detailed report of all dates in season"
        "-vvvvvv: detailed report of all dates in season with missing pitch_app_ids"
    ),
)
@click.pass_obj
def status_season(app, year, verbosity):
    """Report status for a single MLB season."""
    report_type = StatusReport.NONE
    if verbosity <= 0:
        error = f"Invalid value for verbosity: {verbosity}. Value must be greater than zero."
        return exit_app(app, Result.Fail(error))
    elif verbosity == 1:
        report_type = StatusReport.SEASON_SUMMARY
    elif verbosity == 2:
        report_type = StatusReport.DATE_SUMMARY_MISSING_DATA
    elif verbosity == 3:
        report_type = StatusReport.DATE_SUMMARY_ALL_DATES
    elif verbosity == 4:
        report_type = StatusReport.DATE_DETAIL_MISSING_DATA
    elif verbosity == 5:
        report_type = StatusReport.DATE_DETAIL_ALL_DATES
    else:
        report_type = StatusReport.DATE_DETAIL_MISSING_PITCHFX
    result = report_season_status(app.db_session, year, report_type)
    if result.success:
        report_viewer = result.value
        report_viewer.launch()
    return exit_app(app, result)


@cli.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.pass_obj
def sync(app):
    """Synchronize scraped data to/from S3 bucket."""
    pass


@sync.command("up", context_settings={"help_option_names": ["-h", "--help"]})
@click.argument("year", type=MlbSeason(), default=current_year)
@click.option(
    "--file-type",
    type=FileTypeName(),
    help="Type of file to sync, must provide only one value.",
    prompt=True,
)
@click.option(
    "--data-set",
    type=DataSetName(),
    multiple=True,
    default=[str(DataSet.ALL)],
    show_default=True,
    help="Data set(s) to sync, multiple values can be provided.",
)
@click.pass_obj
def sync_up_to_s3(app, year, file_type, data_set):
    """Sync files from local folder to S3 bucket."""
    data_sets_int = sum(int(ds) for ds in flatten_list2d(data_set))
    result_dict = SyncDataNoPromptsTask(app).execute(
        sync_direction=SyncDirection.UP_TO_S3,
        year=year,
        file_type=file_type,
        data_sets_int=data_sets_int,
    )
    result = Result.Combine(list(result_dict.values()))
    if os.environ.get("ENV") != "TEST":  # pragma: no cover
        pause(message="\nPress any key to continue...")
    return exit_app(app, result)


@sync.command("down", context_settings={"help_option_names": ["-h", "--help"]})
@click.argument("year", type=MlbSeason(), default=current_year)
@click.option(
    "--file-type",
    type=FileTypeName(),
    help="Type of file to sync, must provide only one value.",
    prompt=True,
)
@click.option(
    "--data-set",
    type=DataSetName(),
    multiple=True,
    default=[str(DataSet.ALL)],
    show_default=True,
    help="Data set(s) to sync, multiple values can be provided.",
)
@click.pass_obj
def sync_down_to_local(app, year, file_type, data_set):
    """Sync files from S3 bucket to local folder."""
    data_sets_int = sum(int(ds) for ds in flatten_list2d(data_set))
    result_dict = SyncDataNoPromptsTask(app).execute(
        sync_direction=SyncDirection.DOWN_TO_LOCAL,
        year=year,
        file_type=file_type,
        data_sets_int=data_sets_int,
    )
    result = Result.Combine(list(result_dict.values()))
    if os.environ.get("ENV") != "TEST":  # pragma: no cover
        pause(message="\nPress any key to continue...")
    return exit_app(app, result)


def exit_app(app, result, message=None):
    app.db_session.close()
    subprocess.run(["clear"])
    print()
    exit_code = exit_app_success(message) if result.success else exit_app_error(result.error)
    print()
    return exit_code


def exit_app_success(message=None):
    if message:
        print_message(message, fg="bright_green", bold=True)
    return 0


def exit_app_error(message):
    if message:
        if isinstance(message, list):
            for m in message:
                print_message(m, fg="bright_red", bold=True)
        else:
            print_message(message, fg="bright_red", bold=True)
    return 1



import subprocess

from getch import pause
from halo import Halo

from vigorish.cli.components.prompts import user_options_prompt, yes_no_prompt
from vigorish.cli.components.util import (
    get_random_cli_color,
    get_random_dots_spinner,
    print_message,
)
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS, MENU_NUMBERS
from vigorish.tasks import AddToDatabaseTask
from vigorish.util.result import Result


class AddToDatabase(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.add_to_db = AddToDatabaseTask(app)
        self.menu_item_text = "Add Combined Game Data to Database"
        self.menu_item_emoji = EMOJIS.get("BASEBALL")
        self.exit_menu = False
        self.spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.game_ids = []

    @property
    def total_games(self):
        return len(self.game_ids)

    def launch(self):
        if not self.prompt_user_run_task():
            return Result.Ok(True)
        result = self.select_season_prompt()
        if result.failure:
            return Result.Ok()
        year = result.value
        self.subscribe_to_events()
        result = self.add_to_db.execute(year)
        self.spinner.stop()
        self.unsubscribe_from_events()
        if result.failure:
            return result
        subprocess.run(["clear"])
        print_message(f"Added PitchFX data for {len(self.game_ids)} games to database!")
        pause(message="\nPress any key to continue...")
        return Result.Ok()

    def prompt_user_run_task(self):
        task_description = [
            "This task identifies all games that fulfill the two requirements below:\n",
            "1. All data sets have been scraped",
            "2. Data was successfully combined (does not need to be error free)\n",
            "For all games that meet these requirements, the data listed below will be added to ",
            "the database:\n",
            "1. Individual player batting stats",
            "2. Individual player pitching stats",
            "3. PitchFX data for all pitches thrown",
        ]
        subprocess.run(["clear"])
        for line in task_description:
            print_message(line, fg="bright_yellow")
        return yes_no_prompt("\nWould you like to run this task?", wrap=False)

    def select_season_prompt(self):
        prompt = "Select an MLB season from the list below:"
        choices = {f"{MENU_NUMBERS.get(1)}  ALL": "ALL"}
        for num, (year, results) in enumerate(self.app.audit_report.items(), start=2):
            if results["successful"]:
                choices[f"{MENU_NUMBERS.get(num, str(num))}  {year}"] = year
        choices[f"{EMOJIS.get('BACK')} Return to Previous Menu"] = None
        result = user_options_prompt(choices, prompt)
        if result.failure:
            return result
        year = result.value if isinstance(result.value, int) else None
        return Result.Ok(year)

    def add_data_to_db_start(self, year, game_ids):
        subprocess.run(["clear"])
        self.game_ids = game_ids
        self.spinner.text = self.get_progress_text(0, year, game_ids[0])
        self.spinner.start()

    def add_data_to_db_progress(self, num_complete, year, game_id):
        self.spinner.text = self.get_progress_text(num_complete, year, game_id)

    def get_progress_text(self, num_complete, year, game_id):
        percent = num_complete / float(self.total_games)
        return f"Adding combined game data for MLB {year} to database... (Game ID: {game_id}) {percent:.0%}..."

    def subscribe_to_events(self):
        self.add_to_db.events.add_data_to_db_start += self.add_data_to_db_start
        self.add_to_db.events.add_data_to_db_progress += self.add_data_to_db_progress

    def unsubscribe_from_events(self):
        self.add_to_db.events.add_data_to_db_start -= self.add_data_to_db_start
        self.add_to_db.events.add_data_to_db_progress -= self.add_data_to_db_progress



from sqlalchemy import and_, func, join, select
from sqlalchemy_utils.view import create_view

import vigorish.database as db
from vigorish.data.metrics import PitchStatsMetrics
from vigorish.views.pitch_stats_col_expressions import (
    bb_per_nine,
    bb_rate,
    era,
    hr_per_fb,
    hr_per_nine,
    innings_pitched,
    k_minus_bb,
    k_per_bb,
    k_per_nine,
    k_rate,
    whip,
)


class Team_PitchStats_By_Year_View(db.Base):
    __table__ = create_view(
        name="team_pitchstats_by_year",
        selectable=select(
            [
                db.PitchStats.player_team_id.label("team_id"),
                db.PitchStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                func.count(db.PitchStats.game_status_id.distinct()).label("total_games"),
                func.sum(db.PitchStats.is_sp).label("games_as_sp"),
                func.sum(db.PitchStats.is_rp).label("games_as_rp"),
                func.sum(db.PitchStats.is_wp).label("wins"),
                func.sum(db.PitchStats.is_lp).label("losses"),
                func.sum(db.PitchStats.is_sv).label("saves"),
                innings_pitched,
                func.sum(db.PitchStats.total_outs).label("total_outs"),
                func.sum(db.PitchStats.batters_faced).label("batters_faced"),
                func.sum(db.PitchStats.runs).label("runs"),
                func.sum(db.PitchStats.earned_runs).label("earned_runs"),
                func.sum(db.PitchStats.hits).label("hits"),
                func.sum(db.PitchStats.homeruns).label("homeruns"),
                func.sum(db.PitchStats.strikeouts).label("strikeouts"),
                func.sum(db.PitchStats.bases_on_balls).label("bases_on_balls"),
                era,
                whip,
                k_per_nine,
                bb_per_nine,
                hr_per_nine,
                k_per_bb,
                k_rate,
                bb_rate,
                k_minus_bb,
                hr_per_fb,
                func.sum(db.PitchStats.pitch_count).label("pitch_count"),
                func.sum(db.PitchStats.strikes).label("strikes"),
                func.sum(db.PitchStats.strikes_contact).label("strikes_contact"),
                func.sum(db.PitchStats.strikes_swinging).label("strikes_swinging"),
                func.sum(db.PitchStats.strikes_looking).label("strikes_looking"),
                func.sum(db.PitchStats.ground_balls).label("ground_balls"),
                func.sum(db.PitchStats.fly_balls).label("fly_balls"),
                func.sum(db.PitchStats.line_drives).label("line_drives"),
                func.sum(db.PitchStats.unknown_type).label("unknown_type"),
                func.sum(db.PitchStats.inherited_runners).label("inherited_runners"),
                func.sum(db.PitchStats.inherited_scored).label("inherited_scored"),
                func.sum(db.PitchStats.wpa_pitch).label("wpa_pitch"),
                func.sum(db.PitchStats.re24_pitch).label("re24_pitch"),
                db.PitchStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.PitchStats, db.Season, db.PitchStats.season_id == db.Season.id))
        .group_by(db.PitchStats.season_id)
        .group_by(db.PitchStats.player_team_id)
        .order_by(db.PitchStats.player_team_id)
        .order_by(db.PitchStats.season_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pitch_stats_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return PitchStatsMetrics.from_query_results(results)[0] if results else None

    @classmethod
    def get_pitch_stats_by_year_for_team(cls, db_engine, team_id_bbref):
        s = select([cls]).where(cls.team_id_bbref == team_id_bbref)
        results = db_engine.execute(s).fetchall()
        return PitchStatsMetrics.from_query_results(results) if results else []

    @classmethod
    def get_pitch_stats_for_season_for_all_teams(cls, db_engine, year):
        s = select([cls]).where(cls.year == year)
        results = db_engine.execute(s).fetchall()
        return PitchStatsMetrics.from_query_results(results) if results else []


class Team_PitchStats_SP_By_Year_View(db.Base):
    __table__ = create_view(
        name="team_pitchstats_sp_by_year",
        selectable=select(
            [
                db.PitchStats.player_team_id.label("team_id"),
                db.PitchStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                func.count(db.PitchStats.game_status_id.distinct()).label("total_games"),
                func.sum(db.PitchStats.is_sp).label("games_as_sp"),
                func.sum(db.PitchStats.is_rp).label("games_as_rp"),
                func.sum(db.PitchStats.is_wp).label("wins"),
                func.sum(db.PitchStats.is_lp).label("losses"),
                func.sum(db.PitchStats.is_sv).label("saves"),
                innings_pitched,
                func.sum(db.PitchStats.total_outs).label("total_outs"),
                func.sum(db.PitchStats.batters_faced).label("batters_faced"),
                func.sum(db.PitchStats.runs).label("runs"),
                func.sum(db.PitchStats.earned_runs).label("earned_runs"),
                func.sum(db.PitchStats.hits).label("hits"),
                func.sum(db.PitchStats.homeruns).label("homeruns"),
                func.sum(db.PitchStats.strikeouts).label("strikeouts"),
                func.sum(db.PitchStats.bases_on_balls).label("bases_on_balls"),
                era,
                whip,
                k_per_nine,
                bb_per_nine,
                hr_per_nine,
                k_per_bb,
                k_rate,
                bb_rate,
                k_minus_bb,
                hr_per_fb,
                func.sum(db.PitchStats.pitch_count).label("pitch_count"),
                func.sum(db.PitchStats.strikes).label("strikes"),
                func.sum(db.PitchStats.strikes_contact).label("strikes_contact"),
                func.sum(db.PitchStats.strikes_swinging).label("strikes_swinging"),
                func.sum(db.PitchStats.strikes_looking).label("strikes_looking"),
                func.sum(db.PitchStats.ground_balls).label("ground_balls"),
                func.sum(db.PitchStats.fly_balls).label("fly_balls"),
                func.sum(db.PitchStats.line_drives).label("line_drives"),
                func.sum(db.PitchStats.unknown_type).label("unknown_type"),
                func.sum(db.PitchStats.inherited_runners).label("inherited_runners"),
                func.sum(db.PitchStats.inherited_scored).label("inherited_scored"),
                func.sum(db.PitchStats.wpa_pitch).label("wpa_pitch"),
                func.sum(db.PitchStats.re24_pitch).label("re24_pitch"),
                db.PitchStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.PitchStats, db.Season, db.PitchStats.season_id == db.Season.id))
        .where(db.PitchStats.is_sp == 1)
        .group_by(db.PitchStats.season_id)
        .group_by(db.PitchStats.player_team_id)
        .order_by(db.PitchStats.player_team_id)
        .order_by(db.PitchStats.season_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pitch_stats_for_sp_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return PitchStatsMetrics.from_query_results(results)[0] if results else None

    @classmethod
    def get_pitch_stats_for_sp_by_year_for_team(cls, db_engine, team_id_bbref):
        s = select([cls]).where(cls.team_id_bbref == team_id_bbref)
        results = db_engine.execute(s).fetchall()
        return PitchStatsMetrics.from_query_results(results) if results else []

    @classmethod
    def get_pitch_stats_for_sp_for_season_for_all_teams(cls, db_engine, year):
        s = select([cls]).where(cls.year == year)
        results = db_engine.execute(s).fetchall()
        return PitchStatsMetrics.from_query_results(results) if results else []


class Team_PitchStats_RP_By_Year_View(db.Base):
    __table__ = create_view(
        name="team_pitchstats_rp_by_year",
        selectable=select(
            [
                db.PitchStats.player_team_id.label("team_id"),
                db.PitchStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                func.count(db.PitchStats.game_status_id.distinct()).label("total_games"),
                func.sum(db.PitchStats.is_sp).label("games_as_sp"),
                func.sum(db.PitchStats.is_rp).label("games_as_rp"),
                func.sum(db.PitchStats.is_wp).label("wins"),
                func.sum(db.PitchStats.is_lp).label("losses"),
                func.sum(db.PitchStats.is_sv).label("saves"),
                innings_pitched,
                func.sum(db.PitchStats.total_outs).label("total_outs"),
                func.sum(db.PitchStats.batters_faced).label("batters_faced"),
                func.sum(db.PitchStats.runs).label("runs"),
                func.sum(db.PitchStats.earned_runs).label("earned_runs"),
                func.sum(db.PitchStats.hits).label("hits"),
                func.sum(db.PitchStats.homeruns).label("homeruns"),
                func.sum(db.PitchStats.strikeouts).label("strikeouts"),
                func.sum(db.PitchStats.bases_on_balls).label("bases_on_balls"),
                era,
                whip,
                k_per_nine,
                bb_per_nine,
                hr_per_nine,
                k_per_bb,
                k_rate,
                bb_rate,
                k_minus_bb,
                hr_per_fb,
                func.sum(db.PitchStats.pitch_count).label("pitch_count"),
                func.sum(db.PitchStats.strikes).label("strikes"),
                func.sum(db.PitchStats.strikes_contact).label("strikes_contact"),
                func.sum(db.PitchStats.strikes_swinging).label("strikes_swinging"),
                func.sum(db.PitchStats.strikes_looking).label("strikes_looking"),
                func.sum(db.PitchStats.ground_balls).label("ground_balls"),
                func.sum(db.PitchStats.fly_balls).label("fly_balls"),
                func.sum(db.PitchStats.line_drives).label("line_drives"),
                func.sum(db.PitchStats.unknown_type).label("unknown_type"),
                func.sum(db.PitchStats.inherited_runners).label("inherited_runners"),
                func.sum(db.PitchStats.inherited_scored).label("inherited_scored"),
                func.sum(db.PitchStats.wpa_pitch).label("wpa_pitch"),
                func.sum(db.PitchStats.re24_pitch).label("re24_pitch"),
                db.PitchStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.PitchStats, db.Season, db.PitchStats.season_id == db.Season.id))
        .where(db.PitchStats.is_rp == 1)
        .group_by(db.PitchStats.season_id)
        .group_by(db.PitchStats.player_team_id)
        .order_by(db.PitchStats.player_team_id)
        .order_by(db.PitchStats.season_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pitch_stats_for_rp_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return PitchStatsMetrics.from_query_results(results)[0] if results else None

    @classmethod
    def get_pitch_stats_for_rp_by_year_for_team(cls, db_engine, team_id_bbref):
        s = select([cls]).where(cls.team_id_bbref == team_id_bbref)
        results = db_engine.execute(s).fetchall()
        return PitchStatsMetrics.from_query_results(results) if results else []

    @classmethod
    def get_pitch_stats_for_rp_for_season_for_all_teams(cls, db_engine, year):
        s = select([cls]).where(cls.year == year)
        results = db_engine.execute(s).fetchall()
        return PitchStatsMetrics.from_query_results(results) if results else []


class Team_PitchStats_By_Player_By_Year_View(db.Base):
    __table__ = create_view(
        name="team_pitchstats_by_player_by_year",
        selectable=select(
            [
                db.PitchStats.player_team_id.label("team_id"),
                db.PitchStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                db.PitchStats.player_id_mlb.label("mlb_id"),
                db.PitchStats.player_id_bbref.label("bbref_id"),
                func.count(db.PitchStats.game_status_id.distinct()).label("total_games"),
                func.sum(db.PitchStats.is_sp).label("games_as_sp"),
                func.sum(db.PitchStats.is_rp).label("games_as_rp"),
                func.sum(db.PitchStats.is_wp).label("wins"),
                func.sum(db.PitchStats.is_lp).label("losses"),
                func.sum(db.PitchStats.is_sv).label("saves"),
                innings_pitched,
                func.sum(db.PitchStats.total_outs).label("total_outs"),
                func.sum(db.PitchStats.batters_faced).label("batters_faced"),
                func.sum(db.PitchStats.runs).label("runs"),
                func.sum(db.PitchStats.earned_runs).label("earned_runs"),
                func.sum(db.PitchStats.hits).label("hits"),
                func.sum(db.PitchStats.homeruns).label("homeruns"),
                func.sum(db.PitchStats.strikeouts).label("strikeouts"),
                func.sum(db.PitchStats.bases_on_balls).label("bases_on_balls"),
                era,
                whip,
                k_per_nine,
                bb_per_nine,
                hr_per_nine,
                k_per_bb,
                k_rate,
                bb_rate,
                k_minus_bb,
                hr_per_fb,
                func.sum(db.PitchStats.pitch_count).label("pitch_count"),
                func.sum(db.PitchStats.strikes).label("strikes"),
                func.sum(db.PitchStats.strikes_contact).label("strikes_contact"),
                func.sum(db.PitchStats.strikes_swinging).label("strikes_swinging"),
                func.sum(db.PitchStats.strikes_looking).label("strikes_looking"),
                func.sum(db.PitchStats.ground_balls).label("ground_balls"),
                func.sum(db.PitchStats.fly_balls).label("fly_balls"),
                func.sum(db.PitchStats.line_drives).label("line_drives"),
                func.sum(db.PitchStats.unknown_type).label("unknown_type"),
                func.sum(db.PitchStats.inherited_runners).label("inherited_runners"),
                func.sum(db.PitchStats.inherited_scored).label("inherited_scored"),
                func.sum(db.PitchStats.wpa_pitch).label("wpa_pitch"),
                func.sum(db.PitchStats.re24_pitch).label("re24_pitch"),
                db.PitchStats.player_id.label("id"),
                db.PitchStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.PitchStats, db.Season, db.PitchStats.season_id == db.Season.id))
        .group_by(db.PitchStats.season_id)
        .group_by(db.PitchStats.player_id)
        .group_by(db.PitchStats.player_team_id)
        .order_by(db.PitchStats.batters_faced),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pitch_stats_by_player_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return PitchStatsMetrics.from_query_results(results) if results else []


class Team_PitchStats_SP_By_Player_By_Year_View(db.Base):
    __table__ = create_view(
        name="team_pitchstats_for_sp_by_player_by_year",
        selectable=select(
            [
                db.PitchStats.player_team_id.label("team_id"),
                db.PitchStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                db.PitchStats.player_id_mlb.label("mlb_id"),
                db.PitchStats.player_id_bbref.label("bbref_id"),
                func.count(db.PitchStats.game_status_id.distinct()).label("total_games"),
                func.sum(db.PitchStats.is_sp).label("games_as_sp"),
                func.sum(db.PitchStats.is_rp).label("games_as_rp"),
                func.sum(db.PitchStats.is_wp).label("wins"),
                func.sum(db.PitchStats.is_lp).label("losses"),
                func.sum(db.PitchStats.is_sv).label("saves"),
                innings_pitched,
                func.sum(db.PitchStats.total_outs).label("total_outs"),
                func.sum(db.PitchStats.batters_faced).label("batters_faced"),
                func.sum(db.PitchStats.runs).label("runs"),
                func.sum(db.PitchStats.earned_runs).label("earned_runs"),
                func.sum(db.PitchStats.hits).label("hits"),
                func.sum(db.PitchStats.homeruns).label("homeruns"),
                func.sum(db.PitchStats.strikeouts).label("strikeouts"),
                func.sum(db.PitchStats.bases_on_balls).label("bases_on_balls"),
                era,
                whip,
                k_per_nine,
                bb_per_nine,
                hr_per_nine,
                k_per_bb,
                k_rate,
                bb_rate,
                k_minus_bb,
                hr_per_fb,
                func.sum(db.PitchStats.pitch_count).label("pitch_count"),
                func.sum(db.PitchStats.strikes).label("strikes"),
                func.sum(db.PitchStats.strikes_contact).label("strikes_contact"),
                func.sum(db.PitchStats.strikes_swinging).label("strikes_swinging"),
                func.sum(db.PitchStats.strikes_looking).label("strikes_looking"),
                func.sum(db.PitchStats.ground_balls).label("ground_balls"),
                func.sum(db.PitchStats.fly_balls).label("fly_balls"),
                func.sum(db.PitchStats.line_drives).label("line_drives"),
                func.sum(db.PitchStats.unknown_type).label("unknown_type"),
                func.sum(db.PitchStats.inherited_runners).label("inherited_runners"),
                func.sum(db.PitchStats.inherited_scored).label("inherited_scored"),
                func.sum(db.PitchStats.wpa_pitch).label("wpa_pitch"),
                func.sum(db.PitchStats.re24_pitch).label("re24_pitch"),
                db.PitchStats.player_id.label("id"),
                db.PitchStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.PitchStats, db.Season, db.PitchStats.season_id == db.Season.id))
        .where(db.PitchStats.is_sp == 1)
        .group_by(db.PitchStats.season_id)
        .group_by(db.PitchStats.player_id)
        .group_by(db.PitchStats.player_team_id)
        .order_by(db.PitchStats.batters_faced),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pitch_stats_for_sp_by_player_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return PitchStatsMetrics.from_query_results(results) if results else []


class Team_PitchStats_RP_By_Player_By_Year_View(db.Base):
    __table__ = create_view(
        name="team_pitchstats_for_rp_by_player_by_year",
        selectable=select(
            [
                db.PitchStats.player_team_id.label("team_id"),
                db.PitchStats.player_team_id_bbref.label("team_id_bbref"),
                db.Season.year.label("year"),
                db.PitchStats.player_id_mlb.label("mlb_id"),
                db.PitchStats.player_id_bbref.label("bbref_id"),
                func.count(db.PitchStats.game_status_id.distinct()).label("total_games"),
                func.sum(db.PitchStats.is_sp).label("games_as_sp"),
                func.sum(db.PitchStats.is_rp).label("games_as_rp"),
                func.sum(db.PitchStats.is_wp).label("wins"),
                func.sum(db.PitchStats.is_lp).label("losses"),
                func.sum(db.PitchStats.is_sv).label("saves"),
                innings_pitched,
                func.sum(db.PitchStats.total_outs).label("total_outs"),
                func.sum(db.PitchStats.batters_faced).label("batters_faced"),
                func.sum(db.PitchStats.runs).label("runs"),
                func.sum(db.PitchStats.earned_runs).label("earned_runs"),
                func.sum(db.PitchStats.hits).label("hits"),
                func.sum(db.PitchStats.homeruns).label("homeruns"),
                func.sum(db.PitchStats.strikeouts).label("strikeouts"),
                func.sum(db.PitchStats.bases_on_balls).label("bases_on_balls"),
                era,
                whip,
                k_per_nine,
                bb_per_nine,
                hr_per_nine,
                k_per_bb,
                k_rate,
                bb_rate,
                k_minus_bb,
                hr_per_fb,
                func.sum(db.PitchStats.pitch_count).label("pitch_count"),
                func.sum(db.PitchStats.strikes).label("strikes"),
                func.sum(db.PitchStats.strikes_contact).label("strikes_contact"),
                func.sum(db.PitchStats.strikes_swinging).label("strikes_swinging"),
                func.sum(db.PitchStats.strikes_looking).label("strikes_looking"),
                func.sum(db.PitchStats.ground_balls).label("ground_balls"),
                func.sum(db.PitchStats.fly_balls).label("fly_balls"),
                func.sum(db.PitchStats.line_drives).label("line_drives"),
                func.sum(db.PitchStats.unknown_type).label("unknown_type"),
                func.sum(db.PitchStats.inherited_runners).label("inherited_runners"),
                func.sum(db.PitchStats.inherited_scored).label("inherited_scored"),
                func.sum(db.PitchStats.wpa_pitch).label("wpa_pitch"),
                func.sum(db.PitchStats.re24_pitch).label("re24_pitch"),
                db.PitchStats.player_id.label("id"),
                db.PitchStats.season_id.label("season_id"),
            ]
        )
        .select_from(join(db.PitchStats, db.Season, db.PitchStats.season_id == db.Season.id))
        .where(db.PitchStats.is_rp == 1)
        .group_by(db.PitchStats.season_id)
        .group_by(db.PitchStats.player_id)
        .group_by(db.PitchStats.player_team_id)
        .order_by(db.PitchStats.batters_faced),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pitch_stats_for_rp_by_player_for_team(cls, db_engine, team_id_bbref, year):
        s = select([cls]).where(and_(cls.team_id_bbref == team_id_bbref, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return PitchStatsMetrics.from_query_results(results) if results else []



import re
from string import Template

from lxml import html

from vigorish.scrape.brooks_pitch_logs.models.pitch_log import BrooksPitchLog
from vigorish.util.result import Result

PITCHFX_URL_XPATH = '//a[text()="Get Expanded Tabled Data"]/@href'
PITCH_COUNTS_XPATH = '//div[@class="span9"]//table[2]//tr'

T_PITCHFX_URL = "http://www.brooksbaseball.net/pfxVB/${rel_url}"
T_PITCHER_NAME_XPATH = '//option[@value="${id}"]/text()'
T_INNING_NUM_XPATH = '//div[@class="span9"]//table[2]//tr[${index}]/td[1]/text()'
T_PITCH_COUNT_XPATH = '//div[@class="span9"]//table[2]//tr[${index}]/td[2]/text()'

TEAM_ID_PATTERN = r"\((?P<team_id>[\w]{3})\)"
TEAM_ID_REGEX = re.compile(TEAM_ID_PATTERN)


def parse_pitch_log(scraped_html, game, pitcher_id, url):
    page_content = html.fromstring(scraped_html, base_url=url)
    pitch_log = _initialize_pitch_log(game, pitcher_id, url)
    result = _parse_pitcher_details(page_content, game, pitcher_id)
    if result.failure:
        return Result.Ok(pitch_log)
    pitcher_dict = result.value
    pitch_log.pitcher_name = pitcher_dict["name"]
    pitch_log.pitcher_team_id_bb = pitcher_dict["team_id"]
    pitch_log.opponent_team_id_bb = pitcher_dict["opponent_id"]

    parsed = page_content.xpath(PITCHFX_URL_XPATH)
    if not parsed:
        return Result.Ok(pitch_log)
    rel_url = parsed[0]
    pitch_log.pitchfx_url = Template(T_PITCHFX_URL).substitute(rel_url=rel_url)

    result = _parse_pitch_counts(page_content)
    if result.failure:
        return Result.Ok(pitch_log)
    pitch_log.pitch_count_by_inning = result.value

    total_pitches = sum(pitch_log.pitch_count_by_inning.values())
    pitch_log.total_pitch_count = int(total_pitches)
    pitch_log.parsed_all_info = True

    return Result.Ok(pitch_log)


def _initialize_pitch_log(game, pitcher_id, url):
    pitch_log = BrooksPitchLog()
    pitch_log.parsed_all_info = False
    pitch_log.pitcher_id_mlb = int(pitcher_id)
    pitch_log.pitch_app_id = f"{game.bbref_game_id}_{pitcher_id}"
    pitch_log.bb_game_id = game.bb_game_id
    pitch_log.bbref_game_id = game.bbref_game_id
    pitch_log.game_date_year = game.game_date_year
    pitch_log.game_date_month = game.game_date_month
    pitch_log.game_date_day = game.game_date_day
    pitch_log.game_time_hour = game.game_time_hour
    pitch_log.game_time_minute = game.game_time_minute
    pitch_log.time_zone_name = game.time_zone_name
    pitch_log.pitch_log_url = url
    pitch_log.pitcher_name = ""
    pitch_log.pitcher_team_id_bb = ""
    pitch_log.opponent_team_id_bb = ""
    pitch_log.pitchfx_url = ""
    pitch_log.pitch_count_by_inning = {}
    pitch_log.total_pitch_count = 0
    return pitch_log


def _parse_pitcher_details(page_content, game, pitcher_id):
    query = Template(T_PITCHER_NAME_XPATH).substitute(id=pitcher_id)
    parsed = page_content.xpath(query)
    if not parsed:
        error = "Failed to parse pitcher name from game log page."
        return Result.Fail(error)

    selected_pitcher = parsed[0]
    indices = [n for n in range(len(selected_pitcher)) if selected_pitcher.find("-", n) == n]
    if not indices or len(indices) < 2:
        error = "Failed to parse pitcher name from game log page."
        return Result.Fail(error)

    indices.reverse()
    name = selected_pitcher[: indices[1]].strip()
    result = _parse_team_ids(game, selected_pitcher)
    if result.failure:
        return result
    id_dict = result.value
    pitcher_dict = {
        "name": name,
        "team_id": id_dict["team_id"],
        "opponent_id": id_dict["opponent_id"],
    }
    return Result.Ok(pitcher_dict)


def _parse_team_ids(game, selected_pitcher):
    match = TEAM_ID_REGEX.search(selected_pitcher)
    if not match:
        error = "Unable to parse team ids from game log page."
        return Result.Fail(error)
    match_dict = match.groupdict()
    team_id = match_dict["team_id"]
    if team_id.upper() == game.away_team_id_bb:
        id_dict = {"team_id": game.away_team_id_bb, "opponent_id": game.home_team_id_bb}
    else:
        id_dict = {"team_id": game.home_team_id_bb, "opponent_id": game.away_team_id_bb}
    return Result.Ok(id_dict)


def _parse_pitch_counts(page_content):
    rows = page_content.xpath(PITCH_COUNTS_XPATH)
    if not rows:
        error = "Failed to parse inning-by-inning pitch counts from game log page."
        return Result.Fail(error)
    pitch_totals = {}
    for i in range(2, len(rows)):
        q1 = Template(T_INNING_NUM_XPATH).substitute(index=(i + 1))
        q2 = Template(T_PITCH_COUNT_XPATH).substitute(index=(i + 1))
        inning = page_content.xpath(q1)[0]
        int_str = page_content.xpath(q2)[0]
        count = int(int_str, 10)
        pitch_totals[inning] = count
    return Result.Ok(pitch_totals)



from dataclasses import dataclass, field
from typing import List

from vigorish.scrape.bbref_boxscores.models.bat_stats import BBRefBatStats
from vigorish.scrape.bbref_boxscores.models.pitch_stats import BBRefPitchStats
from vigorish.scrape.bbref_boxscores.models.starting_lineup_slot import BBRefStartingLineupSlot
from vigorish.util.list_helpers import as_dict_list


@dataclass
class BBRefBoxscoreTeamData:
    team_id_br: str = ""
    total_wins_before_game: int = 0
    total_losses_before_game: int = 0
    total_runs_scored_by_team: int = 0
    total_runs_scored_by_opponent: int = 0
    total_hits_by_team: int = 0
    total_hits_by_opponent: int = 0
    total_errors_by_team: int = 0
    total_errors_by_opponent: int = 0
    team_won: bool = field(init=False)
    pitcher_of_record: str = ""
    pitcher_earned_save: str = ""
    starting_lineup: List[BBRefStartingLineupSlot] = field(default_factory=list)
    batting_stats: List[BBRefBatStats] = field(default_factory=list)
    pitching_stats: List[BBRefPitchStats] = field(default_factory=list)

    def __post_init__(self):
        self.team_won = self.total_runs_scored_by_team > self.total_runs_scored_by_opponent

    def as_dict(self):
        return {
            "__bbref_boxscore_team_data__": True,
            "team_id_br": self.team_id_br,
            "total_wins_before_game": self.total_wins_before_game,
            "total_losses_before_game": self.total_losses_before_game,
            "total_runs_scored_by_team": self.total_runs_scored_by_team,
            "total_runs_scored_by_opponent": self.total_runs_scored_by_opponent,
            "total_hits_by_team": self.total_hits_by_team,
            "total_hits_by_opponent": self.total_hits_by_opponent,
            "total_errors_by_team": self.total_errors_by_team,
            "total_errors_by_opponent": self.total_errors_by_opponent,
            "team_won": self.team_won,
            "pitcher_of_record": self.pitcher_of_record,
            "pitcher_earned_save": self.pitcher_earned_save,
            "starting_lineup": as_dict_list(self.starting_lineup),
            "batting_stats": as_dict_list(self.batting_stats),
            "pitching_stats": as_dict_list(self.pitching_stats),
        }



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List

from dacite import from_dict
from sqlalchemy.engine import RowProxy

from vigorish.enums import DefensePosition
from vigorish.types import MetricsDict, RowDict
from vigorish.util.dataclass_helpers import get_field_types


@dataclass
class BatStatsMetrics:
    year: int = field(repr=False, default=0)
    team_id_bbref: str = field(repr=False, default="")
    opponent_team_id_bbref: str = field(repr=False, default="")
    is_starter: bool = field(repr=False, default=False)
    bat_order: int = field(repr=False, default=0)
    def_position: DefensePosition = field(repr=False, default=DefensePosition.NONE)
    mlb_id: int = field(repr=False, default=0)
    bbref_id: str = field(repr=False, default="")
    stint_number: int = field(repr=False, default=0)
    total_games: int = field(repr=False, default=0)
    avg: float = 0.0
    obp: float = field(repr=False, default=0.0)
    slg: float = field(repr=False, default=0.0)
    ops: float = 0.0
    iso: float = field(repr=False, default=0.0)
    bb_rate: float = field(repr=False, default=0.0)
    k_rate: float = field(repr=False, default=0.0)
    contact_rate: float = field(repr=False, default=0.0)
    plate_appearances: int = 0
    at_bats: int = field(repr=False, default=0)
    hits: int = field(repr=False, default=0)
    runs_scored: int = field(repr=False, default=0)
    rbis: int = field(repr=False, default=0)
    bases_on_balls: int = field(repr=False, default=0)
    strikeouts: int = field(repr=False, default=0)
    doubles: int = field(repr=False, default=0)
    triples: int = field(repr=False, default=0)
    homeruns: int = field(repr=False, default=0)
    stolen_bases: int = field(repr=False, default=0)
    caught_stealing: int = field(repr=False, default=0)
    hit_by_pitch: int = field(repr=False, default=0)
    intentional_bb: int = field(repr=False, default=0)
    gdp: int = field(repr=False, default=0)
    sac_fly: int = field(repr=False, default=0)
    sac_hit: int = field(repr=False, default=0)
    total_pitches: int = field(repr=False, default=0)
    total_strikes: int = field(repr=False, default=0)
    wpa_bat: float = field(repr=False, default=0.0)
    wpa_bat_pos: float = field(repr=False, default=0.0)
    wpa_bat_neg: float = field(repr=False, default=0.0)
    re24_bat: float = field(repr=False, default=0.0)

    @classmethod
    def from_query_results(cls, results: List[RowProxy]) -> List[BatStatsMetrics]:
        return [from_dict(data_class=cls, data=cls._get_bat_stats(dict(row))) for row in results]

    @classmethod
    def _get_bat_stats(cls, bat_stats: RowDict, decimal_precision=3) -> MetricsDict:
        dc_fields = get_field_types(cls)
        return {
            k: round(v, decimal_precision)
            if dc_fields[k] is float
            else bool(v)
            if dc_fields[k] is bool
            else DefensePosition(int(v))
            if dc_fields[k] is DefensePosition
            else v
            for k, v in bat_stats.items()
            if v and k in dc_fields
        }



from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Union

from vigorish.util.dt_format_strings import DATE_ONLY_2
from vigorish.util.numeric_helpers import ONE_KB


@dataclass
class UrlDetails:
    url: str
    url_id: Union[str, datetime]
    fileName: str
    cachedHtmlFolderPath: str
    scrapedHtmlFolderpath: str

    @property
    def scraped_file_path(self):
        return Path(self.scrapedHtmlFolderpath).joinpath(self.fileName)

    @property
    def scraped_html_is_valid(self):
        return self.scraped_file_path.exists() and self.scraped_file_path.stat().st_size > ONE_KB

    @property
    def scraped_html(self):
        return self.scraped_file_path.read_text() if self.scraped_html_is_valid else None

    @property
    def cached_file_path(self):
        return Path(self.cachedHtmlFolderPath).joinpath(self.fileName)

    @property
    def cached_html_is_valid(self):
        return self.cached_file_path.exists() and self.cached_file_path.stat().st_size > ONE_KB

    @property
    def cached_html(self):
        return self.cached_file_path.read_text() if self.cached_html_is_valid else None

    @property
    def html_was_scraped(self):
        return self.cached_html_is_valid or self.scraped_html_is_valid

    @property
    def html(self):
        return self.cached_html or self.scraped_html or None

    def as_dict(self):
        valid_json_id = self.url_id.strftime(DATE_ONLY_2) if isinstance(self.url_id, datetime) else self.url_id
        return {
            "url": self.url,
            "url_id": valid_json_id,
            "htmlFileName": self.fileName,
            "cachedHtmlFolderPath": self.cachedHtmlFolderPath,
            "htmlFolderpath": self.scrapedHtmlFolderpath,
        }



"""Enum definitions."""
from enum import Enum, IntEnum

from aenum import auto, IntFlag


class VigFile(IntFlag):
    """File types that are scraped, created and stored by vigorish."""

    NONE = 0
    SCRAPED_HTML = auto()
    PARSED_JSON = auto()
    COMBINED_GAME_DATA = auto()
    PATCH_LIST = auto()
    ALL = SCRAPED_HTML | PARSED_JSON | COMBINED_GAME_DATA | PATCH_LIST

    def __str__(self):
        return self.name

    @classmethod
    def from_str(cls, name):
        if name.upper() == "ALL":
            return cls.ALL
        for enum_item in cls:
            if name.upper() == enum_item.name:
                return enum_item
        return cls.NONE


class DataSet(IntFlag):
    """MLB data sets."""

    NONE = 0
    BBREF_GAMES_FOR_DATE = auto()
    BROOKS_GAMES_FOR_DATE = auto()
    BBREF_BOXSCORES = auto()
    BROOKS_PITCH_LOGS = auto()
    BROOKS_PITCHFX = auto()
    ALL = BBREF_GAMES_FOR_DATE | BROOKS_GAMES_FOR_DATE | BBREF_BOXSCORES | BROOKS_PITCH_LOGS | BROOKS_PITCHFX

    def __str__(self):
        return self.name

    @classmethod
    def from_str(cls, name):
        if name.upper() == "ALL":
            return cls.ALL
        for enum_item in cls:
            if name.upper() == enum_item.name:
                return enum_item
        return cls.NONE

    @classmethod
    def get_data_sets_from_int(cls, data_sets_int):
        return [data_set for data_set in cls if data_sets_int & data_set == data_set]


class ConfigType(str, Enum):
    """Data types for configuration settings."""

    STRING = "string"
    NUMERIC = "numeric"
    PATH = "path"
    ENUM = "enum"
    NONE = "none"

    def __str__(self):
        return str.__str__(self)


class ScrapeCondition(str, Enum):
    """Allowed values for SCRAPE_CONDITION config setting."""

    ONLY_MISSING_DATA = "only_missing_data"
    ALWAYS = "always"
    NEVER = "never"

    def __str__(self):
        return str.__str__(self)


class ScrapeTaskOption(str, Enum):
    """Allowed values for SCRAPE_TASK_OPTION config setting."""

    BY_DATE = "by_date"
    BY_DATA_SET = "by_data_set"

    def __str__(self):
        return str.__str__(self)


class HtmlStorageOption(str, Enum):
    """Allowed values for HTML_STORAGE config setting."""

    NONE = "none"
    LOCAL_FOLDER = "local_folder"
    S3_BUCKET = "s3_bucket"
    BOTH = "both"

    def __str__(self):
        return str.__str__(self)


class JsonStorageOption(str, Enum):
    """Allowed values for JSON_STORAGE config setting."""

    LOCAL_FOLDER = "local_folder"
    S3_BUCKET = "s3_bucket"
    BOTH = "both"

    def __str__(self):
        return str.__str__(self)


class CombinedDataStorageOption(str, Enum):
    """Allowed values for COMBINED_DATA_STORAGE config setting."""

    LOCAL_FOLDER = "local_folder"
    S3_BUCKET = "s3_bucket"
    BOTH = "both"

    def __str__(self):
        return str.__str__(self)


class StatusReport(IntEnum):
    """The type of status report (if any) to display."""

    NONE = 0
    SEASON_SUMMARY = 1
    DATE_SUMMARY_MISSING_DATA = 2
    DATE_SUMMARY_ALL_DATES = 3
    DATE_DETAIL_MISSING_DATA = 4
    DATE_DETAIL_ALL_DATES = 5
    DATE_DETAIL_MISSING_PITCHFX = 6
    SINGLE_DATE_WITH_GAME_STATUS = 7

    def __str__(self):
        return self.name


class FileTask(str, Enum):
    """Generic file actions that are mapped to LocalFileTask and S3FileTask file actions."""

    STORE_FILE = "store_file"
    GET_FILE = "get_file"
    REMOVE_FILE = "remove_file"

    def __str__(self):
        return str.__str__(self)


class S3FileTask(str, Enum):
    """Type of action to perform on an object in an S3 bucket."""

    UPLOAD = "upload"
    DOWNLOAD = "download"
    DELETE = "delete"
    RENAME = "rename"

    def __str__(self):
        return str.__str__(self)


class LocalFileTask(str, Enum):
    """Type of action to perform on a local file."""

    WRITE_FILE = "write_file"
    READ_FILE = "read_file"
    DELETE_FILE = "delete_file"
    DECODE_JSON = "decode_file"

    def __str__(self):
        return str.__str__(self)


class JobStatus(IntEnum):
    """Status of a scrape job created by a user."""

    INCOMPLETE = 1
    ERROR = 2
    COMPLETE = 3

    def __str__(self):
        return self.name


class TeamID(str, Enum):
    """Team IDs taken from baseball-reference.com"""

    ARI = "ARI"
    ATL = "ATL"
    BAL = "BAL"
    BOS = "BOS"
    CHW = "CHW"
    CHC = "CHC"
    CIN = "CIN"
    CLE = "CLE"
    COL = "COL"
    DET = "DET"
    HOU = "HOU"
    KCR = "KCR"
    LAA = "LAA"
    LAD = "LAD"
    MIA = "MIA"
    MIL = "MIL"
    MIN = "MIN"
    NYY = "NYY"
    NYM = "NYM"
    OAK = "OAK"
    PHI = "PHI"
    PIT = "PIT"
    SDP = "SDP"
    SEA = "SEA"
    SFG = "SFG"
    STL = "STL"
    TBR = "TBR"
    TEX = "TEX"
    TOR = "TOR"
    WSN = "WSN"

    def __str__(self):
        return self.name


class DefensePosition(IntEnum):
    """Defensive positions and position numbers used when scoring a game."""

    NONE = 0
    PITCHER = 1
    CATCHER = 2
    FIRST_BASE = 3
    SECOND_BASE = 4
    THIRD_BASE = 5
    SHORT_STOP = 6
    LEFT_FIELD = 7
    CENTER_FIELD = 8
    RIGHT_FIELD = 9
    DH = 10
    BENCH = 11
    PINCH_HITTER = 12
    PINCH_RUNNER = 13

    def __str__(self):
        abbrev_dict = {
            "NONE": "NA",
            "PITCHER": "P",
            "CATCHER": "C",
            "FIRST_BASE": "1B",
            "SECOND_BASE": "2B",
            "THIRD_BASE": "3B",
            "SHORT_STOP": "SS",
            "LEFT_FIELD": "LF",
            "CENTER_FIELD": "CF",
            "RIGHT_FIELD": "RF",
            "BENCH": "BN",
            "PINCH_HITTER": "PH",
            "PINCH_RUNNER": "PR",
        }
        return abbrev_dict.get(self.name, self.name)

    @property
    def is_starter(self):
        return int(self) in range(1, 11)

    @classmethod
    def from_abbrev(cls, abbrev):
        for enum_item in cls:
            if abbrev.upper() == str(enum_item):
                return enum_item
        return cls.NONE

    @classmethod
    def from_name(cls, name):
        for enum_item in cls:
            if name.upper() == enum_item.name:
                return enum_item
        return cls.NONE


class PitchType(IntFlag):
    NONE = 0
    CHANGEUP = auto()
    CURVEBALL = auto()
    EEPHUS = auto()
    FASTBALL = auto()
    CUTTER = auto()
    FOUR_SEAM_FASTBALL = auto()
    SPLITTER = auto()
    TWO_SEAM_FASTBALL = auto()
    FORKBALL = auto()
    INTENT_BALL = auto()
    KNUCKLE_BALL_CURVE = auto()
    KNUCKLE_BALL = auto()
    PITCH_OUT = auto()
    SCREWBALL = auto()
    SINKER = auto()
    SLIDER = auto()
    UNKNOWN = auto()
    ALL = (
        CHANGEUP
        | CURVEBALL
        | EEPHUS
        | FASTBALL
        | CUTTER
        | FOUR_SEAM_FASTBALL
        | SPLITTER
        | TWO_SEAM_FASTBALL
        | FORKBALL
        | INTENT_BALL
        | KNUCKLE_BALL_CURVE
        | KNUCKLE_BALL
        | PITCH_OUT
        | SCREWBALL
        | SINKER
        | SLIDER
        | UNKNOWN
    )
    PERCENTILES = (
        CHANGEUP
        | CURVEBALL
        | FASTBALL
        | CUTTER
        | FOUR_SEAM_FASTBALL
        | SPLITTER
        | TWO_SEAM_FASTBALL
        | KNUCKLE_BALL_CURVE
        | SINKER
        | SLIDER
    )

    def __str__(self):
        abbrev_dict = {
            "NONE": "N/A",
            "CHANGEUP": "CH",
            "CURVEBALL": "CU",
            "EEPHUS": "EP",
            "FASTBALL": "FA",
            "CUTTER": "FC",
            "FOUR_SEAM_FASTBALL": "FF",
            "SPLITTER": "FS",
            "TWO_SEAM_FASTBALL": "FT",
            "FORKBALL": "FO",
            "INTENT_BALL": "IN",
            "KNUCKLE_BALL_CURVE": "KC",
            "KNUCKLE_BALL": "KN",
            "PITCH_OUT": "PO",
            "SCREWBALL": "SC",
            "SINKER": "SI",
            "SLIDER": "SL",
            "UNKNOWN": "UN",
        }
        return abbrev_dict.get(self.name, self.name)

    @property
    def print_name(self):
        name_dict = {
            "N/A": "N/A",
            "FF": "Four-seam Fastball",
            "FT": "Two-seam Fastball",
            "IN": "Ball (Intentional)",
            "ALL": "All Pitch Types",
        }
        return name_dict.get(str(self), self.name.replace("_", " ").title())

    @classmethod
    def from_abbrev(cls, abbrev):
        if abbrev.upper() == "ALL":
            return cls.ALL
        for enum_item in cls:
            if abbrev.upper() == str(enum_item):
                return enum_item
        return cls.NONE

    @classmethod
    def from_name(cls, name):
        if name.upper() == "ALL":
            return cls.ALL
        for enum_item in cls:
            if name.upper() == enum_item.name:
                return enum_item
        return cls.NONE

    @classmethod
    def deconstruct_pitch_types_from_int(cls, pitch_mix_int):
        pitch_mix = [pitch_type for pitch_type in cls if pitch_mix_int & pitch_type == pitch_type]
        return [int(pitch_type) for pitch_type in pitch_mix]


class SeasonType(str, Enum):
    """Identifies games, stats, etc as Spring Training, Regular, or Post-Season."""

    PRE_SEASON = "pre_season"
    REGULAR_SEASON = "regular_season"
    POST_SEASON = "post_season"
    NONE = "none"

    def __str__(self):
        season_type_dict = {
            "PRE_SEASON": "Spring Training",
            "REGULAR_SEASON": "Regular Season",
            "POST_SEASON": "Post-Season",
        }
        return season_type_dict.get(self.name, self.name)


class PlayByPlayEvent(str, Enum):
    AT_BAT = "at_bat"
    SUBSTITUTION = "substitution"
    MISC = "misc"

    def __str__(self):
        return str.__str__(self)


class AuditError(str, Enum):
    FAILED_TO_COMBINE = "failed_to_combine"
    PITCHFX_ERROR = "pitchfx_error"
    INVALID_PITCHFX_DATA = "invalid_pitchfx_data"

    def __str__(self):
        error_type_dict = {
            "FAILED_TO_COMBINE": "Failed to combine scraped data",
            "PITCHFX_ERROR": "PitchFX error",
            "INVALID_PITCHFX_DATA": "Invalid PitchFX data",
        }
        return error_type_dict.get(self.name, self.name)


class AuditTask(IntEnum):
    GET_SCRAPED_DATA = 1
    GET_PLAY_BY_PLAY_DATA = 2
    GET_PITCHFX_DATA = 3
    COMBINE_PBP_AND_PFX_DATA = 4
    UPDATE_BOXSCORE = 5

    def __str__(self):
        return self.name


class PatchType(str, Enum):
    CHANGE_BBREF_GAME_ID = "change_bbref_game_id"
    CHANGE_PITCH_SEQUENCE = "change_pitch_sequence"
    CHANGE_BATTER_ID = "change_batter_id"
    CHANGE_PITCHER_ID = "change_pitcher_id"

    def __str__(self):
        return str.__str__(self)


class SyncDirection(str, Enum):
    UP_TO_S3 = "up_to_s3"
    DOWN_TO_LOCAL = "down_to_local"

    def __str__(self):
        return str.__str__(self)



"""Format strings used with strftime/strptime methods."""

DATE_ONLY = "%Y-%m-%d"
DATE_ONLY_2 = "%m/%d/%Y"
DATE_ONLY_TABLE_ID = "%Y%m%d"
DT_NAIVE_LESS = "%m/%d/%Y %I:%M %p"
DT_NAIVE = "%m/%d/%Y %I:%M:%S %p"
DT_AWARE = "%m/%d/%Y %I:%M:%S %p %z"
DT_STR_FORMAT = "%m/%d/%Y %I:%M %p %Z%z"
DT_STR_FORMAT_ALL = "%Y-%m-%d %H:%M:%S.%f %Z%z"
DATE_MONTH_NAME = "%b %d %Y"
FILE_TIMESTAMP = "%Y%m%d_%H%M%S"
CSV_UTC = "%Y-%m-%d %H:%M:%S"
DT_AWARE_VERBOSE = "%A, %B %d %Y %I:%M %p (%Z)"
ISO_8601 = "%Y-%m-%dT%H:%M:%S%z"
HTTP_TIME = "%a, %d %b %Y %H:%M:%S GMT"



import vigorish.database as db
from vigorish.util.result import Result


def update_status_brooks_pitchfx_log_list(scraped_data, db_session, new_pitch_app_ids, apply_patch_list=True):
    pfx_missing_json, pfx_missing_db_records = [], []
    for pitch_app_id in new_pitch_app_ids:
        pitchfx_log = scraped_data.get_brooks_pitchfx_log(pitch_app_id, apply_patch_list)
        if not pitchfx_log:
            pfx_missing_json.append(pitch_app_id)
            continue
        result = update_pitch_appearance_status_records(db_session, pitchfx_log)
        if result.failure:
            pfx_missing_db_records.append(pitch_app_id)
            continue
        db_session.commit()
    return (
        Result.Ok()
        if not (pfx_missing_json or pfx_missing_db_records)
        else Result.Fail(_get_error_message(pfx_missing_json, pfx_missing_db_records))
    )


def _get_error_message(pfx_missing_json, pfx_missing_db_records):
    errors = []
    if pfx_missing_json:
        errors.append(f"Failed to retrieve JSON files for PitchFX logs: {','.join(pfx_missing_json)}")
    if pfx_missing_db_records:
        errors.append(f"pitch_app_id not found in scrape_status_pitch_app table: {','.join(pfx_missing_db_records)}")
    return "\n".join(errors)


def update_pitch_appearance_status_records(db_session, pitchfx_log):
    try:
        pitch_app_status = db.PitchAppScrapeStatus.find_by_pitch_app_id(db_session, pitchfx_log.pitch_app_id)
        if not pitch_app_status:
            return Result.Fail()
        pitch_app_status.scraped_pitchfx = 1
        pitch_app_status.pitch_count_pitchfx = pitchfx_log.total_pitch_count
        return Result.Ok()
    except Exception as e:
        return Result.Fail(f"Error: {repr(e)}")



from collections import defaultdict
from copy import deepcopy

from dacite import from_dict
from events import Events
from tabulate import tabulate

from vigorish.enums import DataSet, PatchType
from vigorish.patch.brooks_pitchfx import (
    BrooksPitchFxPatchList,
    PatchBrooksPitchFxBatterId,
    PatchBrooksPitchFxPitcherId,
)
from vigorish.tasks.base import Task
from vigorish.tasks.combine_scraped_data import CombineScrapedDataTask
from vigorish.util.result import Result
from vigorish.util.string_helpers import validate_at_bat_id

DATA_SET = DataSet.BROOKS_PITCHFX
PATCH_STAT_NAMES = [
    "pitch_count_bbref",
    "pitch_count_pitchfx",
    "batters_faced_bbref",
    "batters_faced_pitchfx",
    "total_at_bats_pitchfx_complete",
    "total_at_bats_patched_pitchfx",
    "total_at_bats_missing_pitchfx",
    "total_at_bats_removed_pitchfx",
    "total_at_bats_pitchfx_error",
    "total_at_bats_invalid_pitchfx",
    "patched_pitchfx_count",
    "missing_pitchfx_count",
    "removed_pitchfx_count",
    "invalid_pitchfx_count",
]


class PatchInvalidPitchFxTask(Task):
    def __init__(self, app):
        super().__init__(app)
        self.combine_data = CombineScrapedDataTask(app)
        self.events = Events(
            (
                "error_occurred",
                "create_invalid_pfx_map_start",
                "create_invalid_pfx_map_complete",
                "match_missing_pfx_data_start",
                "match_missing_pfx_data_complete",
                "create_patch_list_start",
                "create_patch_list_complete",
                "combine_scraped_data_start",
                "combine_scraped_data_complete",
            )
        )

    def execute(self, bbref_game_id, no_prompts=False):
        self.bbref_game_id = bbref_game_id
        self.no_prompts = no_prompts
        result = self.execute_no_prompts() if no_prompts else self.execute_with_prompts()
        if result.failure:
            self.events.error_occurred(result.error)
        return result

    def execute_with_prompts(self):
        self.initialize_attributes()
        result = self.verify_scraped_data_can_be_combined(apply_patch_list=False)
        if result.failure:
            return result
        if not self.game_contains_invalid_pfx():
            return Result.Fail(f"{self.bbref_game_id} does not contain any PitchFX data errors")
        return self.get_invalid_pfx_map()

    def execute_no_prompts(self):
        return (
            self.initialize_attributes()
            .on_success(self.verify_scraped_data_can_be_combined, False)
            .on_success(self.get_invalid_pfx_map)
            .on_success(self.match_missing_pfx_data)
            .on_success(self.create_patch_list)
            .on_success(self.apply_patch_list)
        )

    def initialize_attributes(self):
        self.invalid_pfx_map = {}
        self.game_events = {}
        self.boxscore = {}
        self.match_results = defaultdict(list)
        self.patch_list = None
        return Result.Ok()

    def verify_scraped_data_can_be_combined(self, apply_patch_list):
        results = self.combine_data.investigate(self.bbref_game_id, apply_patch_list)
        if not results["gather_scraped_data_success"]:
            error_message = "Failed to gather all required scaped data for game!"
        elif not results["get_all_pbp_events_success"]:
            error_message = "Failed to process play-by-play events and identify at_bat_ids"
        elif not results["get_all_pfx_data_success"]:
            error_message = "Failed to process PitchFX logs and group by at_bat_id"
        elif not results["combine_data_success"]:
            error_message = "Failed to combine play-by-play and PitchFX data"
        elif not results["update_boxscore_success"]:
            error_message = "Failed to update boxscore with combined play-by-play and PitchFX data"
        else:
            self.game_events = results["game_events_combined_data"]
            self.boxscore = results["boxscore"]
            return Result.Ok()
        return Result.Fail(error_message)

    def game_contains_invalid_pfx(self):
        return self.boxscore["pitchfx_vs_bbref_audit"]["invalid_pitchfx"]

    def game_contains_pfx_errors(self):
        return self.boxscore["pitchfx_vs_bbref_audit"]["pitchfx_error"]

    def game_does_not_contain_any_errors(self):
        return not (self.game_contains_pfx_errors() or self.game_contains_invalid_pfx())

    def get_invalid_pfx_map(self):
        if not self.game_contains_invalid_pfx():
            return Result.Ok()
        self.events.create_invalid_pfx_map_start()
        for inning_id, ab_dict in self.boxscore["invalid_pitchfx"].items():
            for at_bat_id, at_bat_data in ab_dict.items():
                pitch_app_id = at_bat_data["pitch_app_id"]
                if pitch_app_id not in self.invalid_pfx_map:
                    self.invalid_pfx_map[pitch_app_id] = defaultdict(dict)
                self.invalid_pfx_map[pitch_app_id][inning_id][at_bat_id] = {
                    "at_bat_id": at_bat_id,
                    "inning_id": inning_id,
                    "pitch_app_id": pitch_app_id,
                    "at_bat_data": at_bat_data,
                }
        self.events.create_invalid_pfx_map_complete(self.invalid_pfx_map)
        return Result.Ok()

    def match_missing_pfx_data(self):
        if not self.game_contains_invalid_pfx():
            return Result.Ok()
        self.events.match_missing_pfx_data_start()
        for pfx in self.get_invalid_pfx_dict_list():
            matches = self.find_game_events_matching_this_pfx(pfx)
            if not matches:
                match_dict = {"success": False, "invalid_pfx": pfx, "missing_pfx": {}}
                self.match_results["no_matches"].append(match_dict)
                continue
            if len(matches) > 1:
                result = self.check_for_exact_match(pfx)
                if result.failure:
                    match_dict = {"success": False, "invalid_pfx": pfx, "missing_pfx": matches}
                    self.match_results["many_matches"].append(match_dict)
                    continue
                exact_match = result.value
            else:
                exact_match = matches[0]
            match_dict = self.found_successful_match(pfx, exact_match)
            self.match_results["success"].append(match_dict)
        self.events.match_missing_pfx_data_complete(self.match_results)
        return Result.Ok()

    def get_invalid_pfx_dict_list(self):
        return [
            {
                "source": "invalid_pfx",
                "at_bat_id": at_bat_id,
                "inning_id": inning_id[-5:],
                "pitcher_id": ab_dict["at_bat_data"]["pitcher_id_mlb"],
                "pitcher": self.get_pitcher_name_and_team(ab_dict["at_bat_data"]),
                "batter_id": ab_dict["at_bat_data"]["batter_id_mlb"],
                "batter": self.get_batter_name_and_team(ab_dict["at_bat_data"]),
                "pitch_count": len(ab_dict["at_bat_data"]["pitchfx"]),
                "missing_pfx": [],
                "invalid_pfx": sorted(pfx["ab_count"] for pfx in ab_dict["at_bat_data"]["pitchfx"]),
            }
            for pitch_app_dict in self.invalid_pfx_map.values()
            for inning_id, inning_dict in pitch_app_dict.items()
            for at_bat_id, ab_dict in inning_dict.items()
        ]

    def get_pitcher_name_and_team(self, at_bat_data):
        ab_dict = validate_at_bat_id(at_bat_data["at_bat_id"]).value
        return f'{at_bat_data["pitcher_name"]} ({ab_dict["pitcher_team"]})'

    def get_batter_name_and_team(self, at_bat_data):
        ab_dict = validate_at_bat_id(at_bat_data["at_bat_id"]).value
        return f'{at_bat_data["batter_name"]} ({ab_dict["batter_team"]})'

    def find_game_events_matching_this_pfx(self, pfx):
        return [
            self.get_event_dict(event)
            for event in self.game_events
            if event["at_bat_pitchfx_audit"]["missing_pitchfx_count"] > 0 and self.event_matches_pfx(event, pfx)
        ]

    def get_event_dict(self, event):
        ab_dict = validate_at_bat_id(event["at_bat_id"]).value
        return {
            "source": "valid_at_bat",
            "at_bat_id": event["at_bat_id"],
            "inning_id": event["inning_id"][-5:],
            "pitcher_id": event["pitcher_id_mlb"],
            "pitcher": f'{event["pitcher_name"]} ({ab_dict["pitcher_team"]})',
            "batter_id": event["batter_id_mlb"],
            "batter": f'{event["batter_name"]} ({ab_dict["batter_team"]})',
            "pitch_count": event["at_bat_pitchfx_audit"]["pitch_count_bbref"],
            "missing_pfx": event["at_bat_pitchfx_audit"]["missing_pitch_numbers"],
            "invalid_pfx": [],
        }

    def event_matches_pfx(self, event, pfx):
        return (
            event["inning_id"][-5:] == pfx["inning_id"]
            and (event["batter_id_mlb"] == pfx["batter_id"] or event["pitcher_id_mlb"] == pfx["pitcher_id"])
            and event["at_bat_pitchfx_audit"]["missing_pitchfx_count"] <= pfx["pitch_count"]
            and all(p_num in pfx["invalid_pfx"] for p_num in event["at_bat_pitchfx_audit"]["missing_pitch_numbers"])
        )

    def check_for_exact_match(self, pfx):
        # given the invalid pfx data passed as an argument, the process to find an exact match is:
        exact_match = [
            self.get_event_dict(event)
            # iterate through all game events, for each game event:
            for event in self.game_events
            # if game event is missing pitchfx data
            if event["at_bat_pitchfx_audit"]["missing_pitchfx_count"] > 0
            and (
                # AND game_event and invalid pfx took place in the same inning
                event["inning_id"][-5:] == pfx["inning_id"]
                and (
                    # AND game event and invalid pfx have the same batter OR the same pitcher
                    event["batter_id_mlb"] == pfx["batter_id"]
                    or event["pitcher_id_mlb"] == pfx["pitcher_id"]
                )
                # AND number of pitches missing is the same as the number of invalid pfx
                and event["at_bat_pitchfx_audit"]["missing_pitchfx_count"] == pfx["pitch_count"]
                # AND invalid pfx pitch seq. numbers are the same as the missing pitches
                and all(p_num in pfx["invalid_pfx"] for p_num in event["at_bat_pitchfx_audit"]["missing_pitch_numbers"])
            )
        ]
        if not exact_match:
            # zero game events matched all the criteria -> NO EXACT MATCH
            return Result.Fail("")
        if len(exact_match) != 1:
            # more than one game event matched all the criteria -> NO EXACT MATCH
            return Result.Fail("")
        # one game event matched all the criteria -> EXACT MATCH
        return Result.Ok(exact_match[0])

    def found_successful_match(self, pfx, match):
        match_dict = {"success": True, "invalid_pfx": pfx, "missing_pfx": match}
        match_dict["patch_type"] = (
            PatchType.CHANGE_PITCHER_ID
            if match_dict["invalid_pfx"]["pitcher_id"] != match_dict["missing_pfx"]["pitcher_id"]
            else PatchType.CHANGE_BATTER_ID
        )
        match_dict["patch"] = self.no_prompts
        return match_dict

    def create_patch_list(self):
        patch_list = []
        if not self.game_contains_invalid_pfx() or "success" not in self.match_results:
            return Result.Ok()
        self.events.create_patch_list_start()
        for match_dict in self.match_results["success"]:
            if not match_dict["patch"]:
                continue
            result = self.get_game_event(match_dict["missing_pfx"]["at_bat_id"])
            if result.failure:
                return result
            patch_event = result.value
            result = self.get_invalid_pfx(match_dict["invalid_pfx"]["at_bat_id"])
            if result.failure:
                return result
            patch_pfx = result.value
            patches = []
            if match_dict["patch_type"] == PatchType.CHANGE_BATTER_ID:
                patches = self.create_change_batter_id_patch(patch_event, patch_pfx)
            elif match_dict["patch_type"] == PatchType.CHANGE_PITCHER_ID:
                patches = self.create_change_pitcher_id_patch(patch_event, patch_pfx)
            patch_list.extend(patches)
        existing_patch_list = self.scraped_data.get_patch_list(DATA_SET, self.bbref_game_id)
        if existing_patch_list:
            patch_list = self.combine_patch_lists(patch_list, existing_patch_list.patch_list)
        patch_list.sort(key=lambda x: x.park_sv_id)
        self.patch_list = BrooksPitchFxPatchList(patch_list=patch_list, url_id=self.bbref_game_id)
        result = self.scraped_data.save_patch_list(DATA_SET, self.patch_list)
        if result.failure:
            return result
        self.events.create_patch_list_complete(self.patch_list)
        return Result.Ok()

    def get_game_event(self, at_bat_id):
        matches = [event for event in self.game_events if event["at_bat_id"] == at_bat_id]
        if not matches:
            return Result.Fail(f"Error! Failed to retrieve game event for at_bat_id: {at_bat_id}")
        patch_event = matches[0]
        return Result.Ok(patch_event)

    def get_invalid_pfx(self, at_bat_id):
        ab_dict = validate_at_bat_id(at_bat_id).value
        pitch_app_id = ab_dict["pitch_app_id"]
        inning_id = ab_dict["inning_id"]
        patch_pfx = self.invalid_pfx_map[pitch_app_id][inning_id][at_bat_id]["at_bat_data"]
        return (
            Result.Ok(patch_pfx)
            if patch_pfx
            else Result.Fail(f"Error! Failed to retrieve invalid pfx data for at_bat_id: {at_bat_id}")
        )

    def create_change_pitcher_id_patch(self, patch_event, patch_pfx):
        patch_list = []
        for pfx in self.get_all_pfx_data_to_patch(patch_pfx):
            patch = {
                "bbref_game_id": self.bbref_game_id,
                "park_sv_id": pfx["park_sv_id"],
                "current_at_bat_id": patch_pfx["at_bat_id"],
                "current_pitch_app_id": patch_pfx["pitch_app_id"],
                "current_pitcher_id": patch_pfx["pitcher_id_mlb"],
                "current_pitcher_name": patch_pfx["pitcher_name"],
                "new_at_bat_id": patch_event["at_bat_id"],
                "new_pitch_app_id": patch_event["pitch_app_id"],
                "new_pitcher_id": patch_event["pitcher_id_mlb"],
                "new_pitcher_name": patch_event["pitcher_name"],
            }
            patch_list.append(from_dict(data_class=PatchBrooksPitchFxPitcherId, data=patch))
        return patch_list

    def create_change_batter_id_patch(self, patch_event, patch_pfx):
        patch_list = []
        for pfx in self.get_all_pfx_data_to_patch(patch_pfx):
            patch = {
                "bbref_game_id": self.bbref_game_id,
                "park_sv_id": pfx["park_sv_id"],
                "pitch_app_id": patch_pfx["pitch_app_id"],
                "current_at_bat_id": patch_pfx["at_bat_id"],
                "current_batter_id": patch_pfx["batter_id_mlb"],
                "current_batter_name": patch_pfx["batter_name"],
                "new_at_bat_id": patch_event["at_bat_id"],
                "new_batter_id": patch_event["batter_id_mlb"],
                "new_batter_name": patch_event["batter_name"],
            }
            patch_list.append(from_dict(data_class=PatchBrooksPitchFxBatterId, data=patch))
        return patch_list

    def get_all_pfx_data_to_patch(self, invalid_pfx):
        removed_pfx_count = invalid_pfx["at_bat_pitchfx_audit"]["removed_pitchfx_count"]
        return invalid_pfx["pitchfx"] + invalid_pfx["removed_pitchfx"] if removed_pfx_count else invalid_pfx["pitchfx"]

    def combine_patch_lists(self, new_patch_list, old_patch_list):
        new_ids = [p.park_sv_id for p in new_patch_list]
        for p in old_patch_list:
            if p.park_sv_id in new_ids:
                continue
            patch_copy = deepcopy(p)
            new_patch_list.append(patch_copy)
        return new_patch_list

    def apply_patch_list(self):
        if not self.game_contains_invalid_pfx() or not self.patch_list:
            return Result.Ok({"created_patch_list": False})
        self.events.combine_scraped_data_start()
        box_before = deepcopy(self.boxscore)
        result = self.verify_scraped_data_can_be_combined(apply_patch_list=True)
        if result.failure:
            return result
        box_after = deepcopy(self.boxscore)
        patch_results = self.summarize_patch_results(box_before, box_after)
        result = self.combine_data_and_update_db()
        if result.failure:
            return result
        self.events.combine_scraped_data_complete(patch_results)
        return Result.Ok(patch_results)

    def summarize_patch_results(self, boxscore_before, boxscore_after):
        (fixed_all_errors, invalid_pfx, pfx_errors) = self.check_for_remaining_pfx_errors()
        patch_diff_report = self.create_patch_diff_report(boxscore_before, boxscore_after)
        return {
            "created_patch_list": True,
            "patch_diff_report": patch_diff_report,
            "fixed_all_errors": fixed_all_errors,
            "invalid_pfx": invalid_pfx,
            "pfx_errors": pfx_errors,
        }

    def check_for_remaining_pfx_errors(self):
        fixed_all_errors = self.game_does_not_contain_any_errors()
        invalid_pfx = self.game_contains_invalid_pfx()
        pfx_errors = self.game_contains_pfx_errors()
        return (fixed_all_errors, invalid_pfx, pfx_errors)

    def create_patch_diff_report(self, boxscore_before, boxscore_after):
        patch_list = self.scraped_data.get_patch_list(DATA_SET, self.bbref_game_id)
        boxscore_changes = self.compare_pfx_vs_bbref_audit_results(
            audit_before=boxscore_before["pitchfx_vs_bbref_audit"],
            audit_after=boxscore_after["pitchfx_vs_bbref_audit"],
        )
        pitch_stats_before = self.get_pitch_app_stats_dict(boxscore_before)
        pitch_stats_after = self.get_pitch_app_stats_dict(boxscore_after)
        pitch_stat_changes_dict = {}
        for pitch_app_id in self.get_patched_pitch_app_ids(patch_list):
            pitch_stat_changes = self.compare_pfx_vs_bbref_audit_results(
                audit_before=pitch_stats_before[pitch_app_id]["pitch_app_pitchfx_audit"],
                audit_after=pitch_stats_after[pitch_app_id]["pitch_app_pitchfx_audit"],
            )
            pitch_stat_changes_dict[pitch_app_id] = pitch_stat_changes
        return {
            "boxscore_changes": boxscore_changes,
            "pitch_stat_changes_dict": pitch_stat_changes_dict,
            "pitch_stats_after": pitch_stats_after,
        }

    def get_pitch_app_stats_dict(self, boxscore):
        away_team_pitch_stats = boxscore["away_team_data"]["pitching_stats"]
        home_team_pitch_stats = boxscore["home_team_data"]["pitching_stats"]
        all_pitch_stats = away_team_pitch_stats + home_team_pitch_stats
        return {pitch_stats["pitch_app_id"]: pitch_stats for pitch_stats in all_pitch_stats}

    def get_patched_pitch_app_ids(self, patch_list):
        patched_pitch_app_ids = []
        for patch in patch_list.patch_list:
            if isinstance(patch, PatchBrooksPitchFxPitcherId):
                patched_pitch_app_ids.append(patch.current_pitch_app_id)
                patched_pitch_app_ids.append(patch.new_pitch_app_id)
            else:
                patched_pitch_app_ids.append(patch.pitch_app_id)
        return list(set(patched_pitch_app_ids))

    def compare_pfx_vs_bbref_audit_results(self, audit_before, audit_after):
        table_rows = [
            {
                "stat": stat_name,
                "before": audit_before[stat_name],
                "after": audit_after[stat_name],
                "changed": self.get_stat_change(audit_before, audit_after, stat_name),
            }
            for stat_name in PATCH_STAT_NAMES
            if stat_name in audit_before
            and stat_name in audit_after
            and abs(audit_after[stat_name] - audit_before[stat_name]) > 0
        ]
        return tabulate(table_rows, headers="keys")

    def get_stat_change(self, before, after, stat_name):
        change = after[stat_name] - before[stat_name]
        change_sign = "+" if change > 0 else "-" if change < 0 else " "
        return f"{change_sign}{abs(change)}"

    def combine_data_and_update_db(self):
        result = self.combine_data.execute(self.bbref_game_id)
        if not (
            result["gather_scraped_data_success"]
            and result["combined_data_success"]
            and result["update_pitch_apps_success"]
        ):
            # this code should be unreachable
            error = (
                "Critical Error! investigate_errors/combine_data processes produced different "
                f"results for same game: {self.bbref_game_id}"
            )
            return Result.Fail(error)
        return Result.Ok()



import subprocess

from getch import pause
from halo import Halo

from vigorish.cli.components.util import (
    get_random_cli_color,
    get_random_dots_spinner,
    print_heading,
    print_message,
)
from vigorish.cli.components.viewers import DisplayPage, PageViewer
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS
from vigorish.tasks import BackupDatabaseTask
from vigorish.util.result import Result
from vigorish.util.sys_helpers import zip_file_report


class BackupDatabase(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.backup_db = BackupDatabaseTask(self.app)
        self.menu_item_text = " Export Database to CSV"
        self.menu_item_emoji = EMOJIS.get("TABBED_FILES")
        self.exit_menu = False
        self.spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.table_count = 0
        self.total_tables = 0
        self.row_count = 0
        self.total_rows = 0

    def launch(self):
        if not self.prompt_user_run_task():
            return Result.Ok(True)
        self.subscribe_to_events()
        result = self.backup_db.execute()
        self.unsubscribe_from_events()
        self.spinner.stop()
        if result.failure:
            return result
        zip_file = result.value
        subprocess.run(["clear"])
        self.display_zip_file_details(zip_file)
        pause(message="\nPress any key to continue...")
        return Result.Ok(True)

    def prompt_user_run_task(self):
        subprocess.run(["clear"])
        prompt = "Would you like to run this task and backup the database?"
        pages = [DisplayPage(page, None) for page in self.get_task_description_pages()]
        page_viewer = PageViewer(pages, prompt=prompt, text_color="bright_yellow")
        print_heading("Backup Database to CSV", fg="bright_yellow")
        return page_viewer.launch()

    def backup_database_start(self, total_tables):
        subprocess.run(["clear"])
        print_heading("Backup Database to CSV", fg="bright_yellow")
        self.total_tables = total_tables
        self.spinner.start()

    def backup_table_start(self, table_count, table):
        self.table_count = table_count
        self.current_table = table
        self.spinner.text = self.get_spinner_text(0)

    def backup_table_progress(self, percent):
        self.spinner.text = self.get_spinner_text(percent)

    def create_zip_file_start(self):
        self.spinner.text = "Compressing CSV files and adding to ZIP archive..."

    def get_spinner_text(self, percent):
        return (
            f"Exporting {self.current_table.__name__} to CSV... "
            f"(Table {self.table_count}/{self.total_tables}) {percent:.0%}..."
        )

    def display_zip_file_details(self, zip_file):
        print_heading("Backup Database to CSV", fg="bright_yellow", bg=None)
        print_message("Successfully backed up database to .zip file:", bg=None)
        print_message(f"{zip_file}\n", wrap=False)
        print_message("Zipped File Info:\n", fg="bright_cyan", bold=True)
        print_message(zip_file_report(zip_file), fg="bright_cyan", wrap=False)

    def subscribe_to_events(self):
        self.backup_db.events.backup_database_start += self.backup_database_start
        self.backup_db.events.backup_table_start += self.backup_table_start
        self.backup_db.events.backup_table_progress += self.backup_table_progress
        self.backup_db.events.create_zip_file_start += self.create_zip_file_start

    def unsubscribe_from_events(self):
        self.backup_db.events.backup_database_start -= self.backup_database_start
        self.backup_db.events.backup_table_start -= self.backup_table_start
        self.backup_db.events.backup_table_progress -= self.backup_table_progress
        self.backup_db.events.create_zip_file_start -= self.create_zip_file_start

    def get_task_description_pages(self):
        return [
            [
                "This task creates a backup of the database tables listed below:\n",
                "* bat_stats",
                "* pitch_stats",
                "* pitchfx",
                "* scrape_status_date",
                "* scrape_status_game",
                "* scrape_status_pitch_app\n",
                (
                    "After exporting each table as a CSV file, the files are added to a zip "
                    "archive. Then, the zip file is compressed and the CSV files are removed."
                ),
            ],
            [
                (
                    "When resetting/setting up the database, you will be prompted to restore "
                    "these tables if backup files are located based on the value of the "
                    "DB_BACKUP_FOLDER_PATH config setting.\n"
                ),
                (
                    "PLEASE NOTE: The time needed to backup your database depends on the amount "
                    "of data that has been scraped, as well as the compute and memory specs of "
                    "the machine performing the export. For a large database, the time required "
                    "to complete the backup process could take anywhere from 2-10 minutes."
                ),
            ],
        ]



from sqlalchemy import case, cast, Float, func, Integer, String

import vigorish.database as db

innings_whole_ = func.sum(db.PitchStats.total_outs) / 3
innings_whole = cast(cast(innings_whole_, Integer), String)
innings_remainder_ = func.sum(db.PitchStats.total_outs) % 3
innings_remainder = cast(cast(innings_remainder_, Integer), String)
innings_pitched_ = innings_whole + "." + innings_remainder
innings_pitched = cast(innings_pitched_, Float).label("innings_pitched")

era_numerator = func.sum(db.PitchStats.earned_runs) * 27
era_denominator = func.sum(db.PitchStats.total_outs)
era_ = era_numerator / cast(era_denominator, Float)
era = case([(era_denominator > 0, era_)], else_=0.0).label("era")

whip_numerator = (func.sum(db.PitchStats.bases_on_balls) + func.sum(db.PitchStats.hits)) * 3
whip_denominator = func.sum(db.PitchStats.total_outs)
whip_ = whip_numerator / cast(whip_denominator, Float)
whip = case([(whip_denominator > 0, whip_)], else_=0.0).label("whip")

k9_numerator = func.sum(db.PitchStats.strikeouts) * 27
k9_denominator = func.sum(db.PitchStats.total_outs)
k_per_nine_ = k9_numerator / cast(k9_denominator, Float)
k_per_nine = case([(k9_denominator > 0, k_per_nine_)], else_=0.0).label("k_per_nine")

bb9_numerator = func.sum(db.PitchStats.bases_on_balls) * 27
bb9_denominator = func.sum(db.PitchStats.total_outs)
bb_per_nine_ = bb9_numerator / cast(bb9_denominator, Float)
bb_per_nine = case([(bb9_denominator > 0, bb_per_nine_)], else_=0.0).label("bb_per_nine")

hr9_numerator = func.sum(db.PitchStats.homeruns) * 27
hr9_denominator = func.sum(db.PitchStats.total_outs)
hr9_per_nine_ = hr9_numerator / cast(hr9_denominator, Float)
hr_per_nine = case([(hr9_denominator > 0, hr9_per_nine_)], else_=0.0).label("hr_per_nine")

k_per_bb_numerator = func.sum(db.PitchStats.strikeouts)
k_per_bb_denominator = func.sum(db.PitchStats.bases_on_balls)
k_per_bb_ = k_per_bb_numerator / cast(k_per_bb_denominator, Float)
k_per_bb = case([(k_per_bb_denominator > 0, k_per_bb_)], else_=0.0).label("k_per_bb")

k_rate_numerator = func.sum(db.PitchStats.strikeouts)
k_rate_denominator = func.sum(db.PitchStats.batters_faced)
k_rate_ = k_rate_numerator / cast(k_rate_denominator, Float)
k_rate = case([(k_rate_denominator > 0, k_rate_)], else_=0.0).label("k_rate")

bb_rate_numerator = func.sum(db.PitchStats.bases_on_balls)
bb_rate_denominator = func.sum(db.PitchStats.batters_faced)
bb_rate_ = bb_rate_numerator / cast(bb_rate_denominator, Float)
bb_rate = case([(bb_rate_denominator > 0, bb_rate_)], else_=0.0).label("bb_rate")

k_minus_bb = (k_rate_ - bb_rate_).label("k_minus_bb")

hr_per_fb_numerator = func.sum(db.PitchStats.homeruns)
hr_per_fb_denominator = func.sum(db.PitchStats.fly_balls)
hr_per_fb_ = hr_per_fb_numerator / cast(hr_per_fb_denominator, Float)
hr_per_fb = case([(hr_per_fb_denominator > 0, hr_per_fb_)], else_=0.0).label("hr_per_fb")
